# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.operation module

"""

from dataclasses import dataclass, field
from marshmallow import fields
from typing import Union
from marshmallow_dataclass import class_schema
from keria.app import aiding, credentialing
from keria.app.credentialing import (
    Anchor,
    AnchoringEvent,
    ACDC_V_1,
    ACDC_V_2,
    ICP_V_1,
    ICP_V_2,
    ROT_V_1,
    ROT_V_2,
    DIP_V_1,
    DIP_V_2,
    IXN_V_1,
    IXN_V_2,
    DRT_V_1,
    DRT_V_2,
)
from keria.app.aiding import (
    KeyStateRecord,
    RPY_V_1,
    RPY_V_2,
)
from keria.peer.exchanging import EXN_V_1, EXN_V_2
from keria.core.longrunning import (
    OperationStatus,
    PendingOperation,
    CompletedOperation,
    FailedOperation,
)


@dataclass
class OOBIMetadata:
    oobi: str


@dataclass
class BaseOOBIOperation:
    metadata: OOBIMetadata = field(
        default_factory=OOBIMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OOBIMetadata), required=False
            )
        },
    )


@dataclass
class PendingOOBIOperation(BaseOOBIOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedOOBIOperation(BaseOOBIOperation, CompletedOperation):
    response: aiding.KeyStateRecord = field(
        default_factory=aiding.KeyStateRecord,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(aiding.KeyStateRecord), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedOOBIOperation(BaseOOBIOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


OOBIOperation = Union[PendingOOBIOperation, CompletedOOBIOperation, FailedOOBIOperation]


@dataclass
class QueryMetadata:
    pre: str
    sn: int
    anchor: credentialing.Anchor = field(
        default_factory=credentialing.Anchor,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(credentialing.Anchor), required=False
            )
        },
    )


@dataclass
class BaseQueryOperation:
    metadata: QueryMetadata = field(
        default_factory=QueryMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(QueryMetadata), required=False
            )
        },
    )


@dataclass
class PendingQueryOperation(BaseQueryOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedQueryOperation(BaseQueryOperation, CompletedOperation):
    response: aiding.KeyStateRecord = field(
        default_factory=aiding.KeyStateRecord,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(aiding.KeyStateRecord), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedQueryOperation(BaseQueryOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


QueryOperation = Union[
    PendingQueryOperation, CompletedQueryOperation, FailedQueryOperation
]


@dataclass
class WitnessMetadata:
    pre: str
    sn: int


@dataclass
class BaseWitnessOperation:
    metadata: WitnessMetadata = field(
        default_factory=WitnessMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(WitnessMetadata), required=False
            )
        },
    )


@dataclass
class PendingWitnessOperation(BaseWitnessOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedWitnessOperation(BaseWitnessOperation, CompletedOperation):
    response: Union[ICP_V_1, ICP_V_2, ROT_V_1, ROT_V_2, IXN_V_1, IXN_V_2] = field(
        default=None, metadata={"required": True}
    )  # type: ignore


@dataclass(kw_only=True)
class FailedWitnessOperation(BaseWitnessOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


WitnessOperation = Union[
    PendingWitnessOperation, CompletedWitnessOperation, FailedWitnessOperation
]


@dataclass
class DelegationMetadata:
    pre: str
    sn: int


@dataclass
class BaseDelegationOperation:
    metadata: DelegationMetadata = field(
        default_factory=DelegationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(DelegationMetadata), required=False
            )
        },
    )


@dataclass
class PendingDelegationOperation(BaseDelegationOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedDelegationOperation(BaseDelegationOperation, CompletedOperation):
    response: Union[DIP_V_1, DIP_V_2, DRT_V_1, DRT_V_2] = field(
        default=None, metadata={"required": True}
    )  # type: ignore


@dataclass(kw_only=True)
class FailedDelegationOperation(BaseDelegationOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


DelegationOperation = Union[
    PendingDelegationOperation, CompletedDelegationOperation, FailedDelegationOperation
]


@dataclass
class DoneOperationMetadata:
    response: Union[ICP_V_1, ICP_V_2, ROT_V_1, ROT_V_2, EXN_V_1, EXN_V_2]  # type: ignore
    pre: str = None


@dataclass
class BaseDoneOperation:
    metadata: DoneOperationMetadata = field(
        default_factory=DoneOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(DoneOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingDoneOperation(BaseDoneOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedDoneOperation(BaseDoneOperation, CompletedOperation):
    response: Union[ICP_V_1, ICP_V_2, ROT_V_1, ROT_V_2, EXN_V_1, EXN_V_2] = field(
        default=None, metadata={"required": True}
    )  # type: ignore


@dataclass(kw_only=True)
class FailedDoneOperation(BaseDoneOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


DoneOperation = Union[PendingDoneOperation, CompletedDoneOperation, FailedDoneOperation]


@dataclass
class GroupOperationMetadata:
    pre: str
    sn: int


@dataclass
class BaseGroupOperation:
    metadata: GroupOperationMetadata = field(
        default_factory=GroupOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(GroupOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingGroupOperation(BaseGroupOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedGroupOperation(BaseGroupOperation, CompletedOperation):
    response: AnchoringEvent = field(default=None, metadata={"required": True})  # type: ignore


@dataclass(kw_only=True)
class FailedGroupOperation(BaseGroupOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


GroupOperation = Union[
    PendingGroupOperation, CompletedGroupOperation, FailedGroupOperation
]


@dataclass
class DelegatorOperationMetadata:
    pre: str
    teepre: str
    anchor: credentialing.Anchor = field(
        default_factory=credentialing.Anchor,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(credentialing.Anchor), required=False
            )
        },
    )
    depends: Union["GroupOperation", "WitnessOperation", "DoneOperation"] = None  # type: ignore


@dataclass
class BaseDelegatorOperation:
    metadata: DelegatorOperationMetadata = field(
        default_factory=DelegatorOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(DelegatorOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingDelegatorOperation(BaseDelegatorOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedDelegatorOperation(BaseDelegatorOperation, CompletedOperation):
    response: str


@dataclass(kw_only=True)
class FailedDelegatorOperation(BaseDelegatorOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


DelegatorOperation = Union[
    PendingDelegatorOperation, CompletedDelegatorOperation, FailedDelegatorOperation
]


@dataclass
class SubmitOperationMetadata:
    pre: str
    sn: int


@dataclass
class BaseSubmitOperation:
    metadata: SubmitOperationMetadata = field(
        default_factory=SubmitOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(SubmitOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingSubmitOperation(BaseSubmitOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedSubmitOperation(BaseSubmitOperation, CompletedOperation):
    response: KeyStateRecord = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(KeyStateRecord), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedSubmitOperation(BaseSubmitOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


SubmitOperation = Union[
    PendingSubmitOperation, CompletedSubmitOperation, FailedSubmitOperation
]


@dataclass
class EndRoleMetadata:
    cid: str
    role: str
    eid: str


@dataclass
class BaseEndRoleOperation:
    metadata: EndRoleMetadata = field(
        default_factory=EndRoleMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(EndRoleMetadata), required=False
            )
        },
    )


@dataclass
class PendingEndRoleOperation(BaseEndRoleOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedEndRoleOperation(BaseEndRoleOperation, CompletedOperation):
    response: Union[RPY_V_1, RPY_V_2] = field(default=None, metadata={"required": True})  # type: ignore


@dataclass(kw_only=True)
class FailedEndRoleOperation(BaseEndRoleOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


EndRoleOperation = Union[
    PendingEndRoleOperation, CompletedEndRoleOperation, FailedEndRoleOperation
]


@dataclass
class LocSchemeMetadata:
    eid: str
    scheme: str
    url: str


@dataclass
class BaseLocSchemeOperation:
    metadata: LocSchemeMetadata = field(
        default_factory=LocSchemeMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(LocSchemeMetadata), required=False
            )
        },
    )


@dataclass
class PendingLocSchemeOperation(BaseLocSchemeOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedLocSchemeOperation(BaseLocSchemeOperation, CompletedOperation):
    response: LocSchemeMetadata = field(
        default_factory=LocSchemeMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(LocSchemeMetadata), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedLocSchemeOperation(BaseLocSchemeOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


LocSchemeOperation = Union[
    PendingLocSchemeOperation, CompletedLocSchemeOperation, FailedLocSchemeOperation
]


@dataclass
class ChallengeOperationMetadata:
    words: list[str]


@dataclass
class ChallengeOperationResponse:
    exn: Union[EXN_V_1, EXN_V_2]  # type: ignore


@dataclass
class BaseChallengeOperation:
    metadata: ChallengeOperationMetadata = field(
        default_factory=ChallengeOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(ChallengeOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingChallengeOperation(BaseChallengeOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedChallengeOperation(BaseChallengeOperation, CompletedOperation):
    response: ChallengeOperationResponse = field(
        default_factory=ChallengeOperationResponse,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(ChallengeOperationResponse), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedChallengeOperation(BaseChallengeOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


ChallengeOperation = Union[
    PendingChallengeOperation, CompletedChallengeOperation, FailedChallengeOperation
]


@dataclass
class RegistryOperationMetadata:
    pre: str
    depends: Union[
        "GroupOperation", "WitnessOperation", "DoneOperation", "DelegationOperation"
    ]
    anchor: Anchor = field(
        default_factory=Anchor,
        metadata={
            "marshmallow_field": fields.Nested(class_schema(Anchor), required=True)
        },
    )


@dataclass
class RegistryOperationResponse:
    anchor: Anchor = field(
        default_factory=Anchor,
        metadata={
            "marshmallow_field": fields.Nested(class_schema(Anchor), required=True)
        },
    )


@dataclass
class BaseRegistryOperation:
    metadata: RegistryOperationMetadata = field(
        default_factory=RegistryOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(RegistryOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingRegistryOperation(BaseRegistryOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedRegistryOperation(BaseRegistryOperation, CompletedOperation):
    response: RegistryOperationResponse = field(
        default_factory=RegistryOperationResponse,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(RegistryOperationResponse), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedRegistryOperation(BaseRegistryOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


RegistryOperation = Union[
    PendingRegistryOperation, CompletedRegistryOperation, FailedRegistryOperation
]


@dataclass
class CredentialOperationMetadata:
    ced: Union[ACDC_V_1, ACDC_V_2]  # type: ignore
    depends: Union[ROT_V_1, ROT_V_2, DRT_V_1, DRT_V_2, IXN_V_1, IXN_V_2] = None  # type: ignore


@dataclass
class CredentialOperationResponse:
    ced: Union[ACDC_V_1, ACDC_V_2] = None  # type: ignore


@dataclass
class BaseCredentialOperation:
    metadata: CredentialOperationMetadata = field(
        default_factory=CredentialOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(CredentialOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingCredentialOperation(BaseCredentialOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedCredentialOperation(BaseCredentialOperation, CompletedOperation):
    response: CredentialOperationResponse = field(
        default_factory=CredentialOperationResponse,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(CredentialOperationResponse), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedCredentialOperation(BaseCredentialOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


CredentialOperation = Union[
    PendingCredentialOperation, CompletedCredentialOperation, FailedCredentialOperation
]


@dataclass
class ExchangeOperationMetadata:
    said: str


@dataclass
class BaseExchangeOperation:
    metadata: ExchangeOperationMetadata = field(
        default_factory=ExchangeOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(ExchangeOperationMetadata), required=False
            )
        },
    )


@dataclass
class PendingExchangeOperation(BaseExchangeOperation, PendingOperation):
    pass


@dataclass(kw_only=True)
class CompletedExchangeOperation(BaseExchangeOperation, CompletedOperation):
    response: ExchangeOperationMetadata = field(
        default_factory=ExchangeOperationMetadata,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(ExchangeOperationMetadata), required=True
            )
        },
    )


@dataclass(kw_only=True)
class FailedExchangeOperation(BaseExchangeOperation, FailedOperation):
    error: OperationStatus = field(
        default=None,
        metadata={
            "marshmallow_field": fields.Nested(
                class_schema(OperationStatus), required=True
            )
        },
    )


ExchangeOperation = Union[
    PendingExchangeOperation, CompletedExchangeOperation, FailedExchangeOperation
]



# -*- encoding: utf-8 -*-
"""
Test for keria.__init__
"""

import logging
import unittest
from unittest.mock import Mock
import keria


class TestKeriaInit(unittest.TestCase):
    def test_set_log_level_with_debug(self):
        mock_logger = Mock()
        keria.set_log_level("DEBUG", mock_logger)
        self.assertEqual(keria.ogler.level, logging.DEBUG)
        mock_logger.setLevel.assert_called_with(logging.DEBUG)


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, List, Union, Optional
from enum import Enum
from dataclasses import field, make_dataclass
from marshmallow_dataclass import class_schema
from marshmallow import fields as mm_fields
from keri.core import serdering


class OneOfField(mm_fields.Field):
    """Generic field for oneOf schema types that generates proper OpenAPI oneOf schema."""

    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        return value

    def __init__(self, schema_options: List[Dict[str, Any]], **kwargs):
        """
        Initialize OneOfField with schema options.

        Args:
            schema_options: List of OpenAPI schema dictionaries for oneOf
            **kwargs: Additional marshmallow field kwargs
        """
        super().__init__(**kwargs)
        self.metadata["oneOf"] = schema_options


class StringOrStringArrayField(OneOfField):
    def __init__(self, **kwargs):
        super().__init__(
            [{"type": "string"}, {"type": "array", "items": {"type": "string"}}],
            **kwargs,
        )


class StringOrArrayOrArrayOfArraysField(OneOfField):
    def __init__(self, **kwargs):
        super().__init__(
            [
                {"type": "string"},
                {"type": "array", "items": {"type": "string"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "string"}},
                },
            ],
            **kwargs,
        )


class UnionField(mm_fields.Field):
    """Custom field for generic Union types that generates anyOf schema."""

    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        return value

    def __init__(self, union_type, **kwargs):
        super().__init__(**kwargs)
        self.metadata["anyOf"] = []
        for arg in union_type.__args__:
            if arg is str:
                self.metadata["anyOf"].append({"type": "string"})
            elif hasattr(arg, "__origin__") and arg.__origin__ is list:
                if arg.__args__ and arg.__args__[0] is str:
                    self.metadata["anyOf"].append(
                        {"type": "array", "items": {"type": "string"}}
                    )
                else:
                    self.metadata["anyOf"].append({"type": "array"})


def inferType(value: Any, key: str = ""):
    """Infer the Python type from a value, with special handling for certain keys."""
    if key in ("a", "A"):
        return Any
    if isinstance(value, str):
        return str
    elif isinstance(value, int):
        return int
    elif isinstance(value, float):
        return float
    elif isinstance(value, bool):
        return bool
    elif isinstance(value, list):
        if len(value) > 0:
            item_type = inferType(value[0], key)
        else:
            item_type = str
        return List[item_type]
    elif isinstance(value, dict):
        return Dict[str, Any]
    else:
        return Any


def createDataclassNestedField(
    key: str, customType: type, isOptional: bool
) -> tuple[tuple, mm_fields.Field]:
    """Create nested field for dataclass types.

    Handles regular dataclass types like ACDCAttributes, Seal, etc. by creating
    marshmallow Nested fields with proper class_schema() serialization.

    Args:
        key: Field name from the FieldDom
        customType: Dataclass type (must have __dataclass_fields__ attribute)
        isOptional: Whether the field should be optional in the schema

    Returns:
        Tuple of (field_definition_for_dataclass, marshmallow_field_for_schema)
    """
    nestedSchema = class_schema(customType)()
    return createOptionalField(
        key, customType, mm_fields.Nested, (nestedSchema,), {}, isOptional
    )


def createOptionalField(
    key: str,
    fieldType: type,
    marshmallowFieldClass: type,
    marshmallowFieldArgs: tuple = None,
    marshmallowFieldKwargs: dict = None,
    isOptional: bool = True,
) -> tuple[tuple, mm_fields.Field]:
    """
    Create marshmallow field and fieldDef tuple with consistent optional/required handling.

    Args:
        key: Field name
        fieldType: Python type for the field
        marshmallowFieldClass: Marshmallow field class to instantiate
        marshmallowFieldArgs: Positional arguments for marshmallow field constructor
        marshmallowFieldKwargs: Additional kwargs for marshmallow field (beyond required/missing/default/allow_none)
        isOptional: Whether field should be optional

    Returns:
        Tuple of (fieldDef, marshmallowField) where fieldDef is tuple for dataclass field definition
    """
    if marshmallowFieldArgs is None:
        marshmallowFieldArgs = ()
    if marshmallowFieldKwargs is None:
        marshmallowFieldKwargs = {}

    if isOptional:
        marshmallowField = marshmallowFieldClass(
            *marshmallowFieldArgs,
            required=False,
            missing=mm_fields.missing_,
            default=mm_fields.missing_,
            allow_none=False,
            **marshmallowFieldKwargs,
        )
        fieldDef = (key, Optional[fieldType], field(default=None))
    else:
        marshmallowField = marshmallowFieldClass(
            *marshmallowFieldArgs,
            required=True,
            missing=mm_fields.missing_,
            default=mm_fields.missing_,
            **marshmallowFieldKwargs,
        )
        fieldDef = (key, fieldType, field())

    return fieldDef, marshmallowField


def createCustomNestedField(
    key: str, customType: type, isOptional: bool
) -> tuple[tuple, mm_fields.Field]:
    """Create custom nested field for specified type.

    This function is essential for converting KERI protocol FieldDom specifications into proper
    marshmallow fields for OpenAPI schema generation.

    Args:
        key: Field name from the FieldDom
        customType: Type annotation (could be Union, List[DataClass], regular DataClass, etc.)
        isOptional: Whether the field should be optional in the schema

    Returns:
        Tuple of (field_definition_for_dataclass, marshmallow_field_for_schema)
    """
    # Check if this is a Union type (e.g., Union[str, List[str]])
    if hasattr(customType, "__origin__") and customType.__origin__ is Union:
        # Handle Union types - these create polymorphic fields in KERI protocol
        union_args = customType.__args__

        # Check for the specific pattern Union[str, List[str]] which is common in KERI
        # This generates a more specific oneOf schema instead of generic anyOf
        if len(union_args) == 2 and str in union_args:
            # Find if one of the union args is List[str]
            list_type = next(
                (
                    arg
                    for arg in union_args
                    if hasattr(arg, "__origin__") and arg.__origin__ is list
                ),
                None,
            )
            # Ensure it's List[str] specifically (single string type argument)
            # This matches the KERI pattern Union[str, List[str]] exactly
            if (
                list_type
                and len(list_type.__args__) == 1
                and list_type.__args__[0] is str
            ):
                return createOptionalField(
                    key, customType, StringOrStringArrayField, (), {}, isOptional
                )

        # For other Union types, fall back to Raw field with anyOf
        return createOptionalField(
            key, customType, UnionField, (customType,), {}, isOptional
        )

    # Check if this is a generic type like List[SomeClass], Dict[str, Any], etc.
    # __origin__ identifies the base generic type (list, dict, etc.)
    elif hasattr(customType, "__origin__"):
        if customType.__origin__ is list:
            # Handle List[SomeClass] types - common in KERI for repeated elements
            # __args__ contains the type parameters, e.g., List[Seal] has __args__ = (Seal,)
            if customType.__args__:
                inner_type = customType.__args__[0]
                # Check if the inner type is a dataclass (has __dataclass_fields__ attribute)
                # This ensures we can safely call class_schema() on it for nested serialization
                if hasattr(inner_type, "__dataclass_fields__"):
                    # Create a List field with nested schema
                    nestedField = mm_fields.Nested(class_schema(inner_type)())
                    return createOptionalField(
                        key, customType, mm_fields.List, (nestedField,), {}, isOptional
                    )

        # For other generic types, fall back to Raw field
        return createOptionalField(key, customType, mm_fields.Raw, (), {}, isOptional)

    # This validation prevents AttributeError: __mro__ when calling class_schema() on non-dataclass types
    elif hasattr(customType, "__dataclass_fields__"):
        return createDataclassNestedField(key, customType, isOptional)

    # For non-dataclass, non-generic types, use Raw field
    else:
        return createOptionalField(key, customType, mm_fields.Raw, (), {}, isOptional)


def createKtNtField(key: str, isOptional: bool) -> tuple[tuple, mm_fields.Field]:
    # Type can be string, List[str], or List[List[str]]
    pyType = Union[str, List[str], List[List[str]]]

    return createOptionalField(
        key, pyType, StringOrArrayOrArrayOfArraysField, (), {}, isOptional
    )


def createRegularField(
    key: str, value: Any, isOptional: bool = True
) -> tuple[tuple, Optional[mm_fields.Field]]:
    """Create regular field without custom nested handling."""
    inferredType = inferType(value, key)

    marshmallowField = None
    if isOptional and key not in ("a", "A"):
        if inferredType is str or inferredType is Optional[str]:
            marshmallowField = mm_fields.String(
                required=False,
                missing=mm_fields.missing_,
                default=mm_fields.missing_,
                allow_none=False,
            )
        else:
            marshmallowField = mm_fields.Field(
                required=False,
                missing=mm_fields.missing_,
                default=mm_fields.missing_,
                allow_none=False,
            )

    if isOptional:
        fieldDef = (key, Optional[inferredType], field(default=None))
    else:
        fieldDef = (key, inferredType)

    return fieldDef, marshmallowField


def processField(
    key: str,
    value: Any,
    fieldDom: serdering.FieldDom,
    customTypes: Dict[str, type],
    name: str = "",
) -> tuple[tuple, Optional[mm_fields.Field]]:
    """Process a single field from the FieldDom."""
    isOptional = key in fieldDom.opts

    # Check if there's a custom type specified
    if key in customTypes:
        return createCustomNestedField(key, customTypes[key], isOptional)

    # Custom: for "kt" or "nt", allow string or array of strings or array of array of strings
    elif key in ("kt", "nt"):
        return createKtNtField(key, isOptional)

    else:
        return createRegularField(key, value, isOptional)


def dataclassFromFielddom(
    name: str,
    fieldDom: serdering.FieldDom,
    customTypes: Optional[Dict[str, type]] = None,
):
    """
    Generates a dataclass from a FieldDom instance with proper marshmallow schema.
    Args:
        name: Name for the generated dataclass
        fieldDom: FieldDom instance containing field definitions
        customTypes: Optional dict mapping field names to custom types

    Returns:
        Tuple of (generated_dataclass, marshmallow_schema)
    """
    if customTypes is None:
        customTypes = {}

    requiredFields = []
    optionalFields = []
    customFields = {}

    # Store alt constraints for use by schema generation
    altConstraints = getattr(fieldDom, "alts", {})

    # Process all fields from alls (all possible fields)
    for key, value in fieldDom.alls.items():
        fieldDef, marshmallowField = processField(
            key, value, fieldDom, customTypes, name
        )

        if marshmallowField:
            customFields[key] = marshmallowField

        # Check if field is optional (in opts)
        isOptional = key in fieldDom.opts
        if isOptional:
            optionalFields.append(fieldDef)
        else:
            requiredFields.append(fieldDef)

    allFields = requiredFields + optionalFields
    generatedCls = make_dataclass(name, allFields)
    schema = class_schema(generatedCls)()

    # Override the automatically generated fields with our custom ones
    # marshmallow_dataclass automatically creates fields with allow_none=True for Optional[T]
    # We need to override them to have allow_none=False for openapi-typescript compatibility
    for key, marshmallowField in customFields.items():
        schema.fields[key] = marshmallowField
        schema._declared_fields[key] = marshmallowField

    # Store alt constraints for later use in OpenAPI generation
    schema._alt_constraints = altConstraints

    return generatedCls, schema


def applyAltConstraintsToOpenApiSchema(
    openApiSchemaDict: dict, altConstraints: Dict[str, str]
):
    """
    Modify an OpenAPI schema to include oneOf constraints for alternate fields.

    This should be called after the schema is registered in the OpenAPI spec.

    Args:
        openApiSchemaDict: The OpenAPI schema dictionary to modify
        altConstraints: Dict mapping alternate fields (e.g., {'a': 'A', 'A': 'a'})
    """
    if not altConstraints or "properties" not in openApiSchemaDict:
        return

    properties = openApiSchemaDict.get("properties", {})
    required = openApiSchemaDict.get("required", [])

    # Find alternate field pairs that exist in properties
    altGroups = {}
    processedAlts = set()

    for field1, field2 in altConstraints.items():
        if field1 in processedAlts or field2 in processedAlts:
            continue
        if field1 in properties and field2 in properties:
            groupKey = f"{field1}_{field2}"
            altGroups[groupKey] = [field1, field2]
            processedAlts.add(field1)
            processedAlts.add(field2)

    if not altGroups:
        return

    # Create oneOf schemas for alternate field combinations
    oneOfSchemas = []

    for _, altFields in altGroups.items():
        field1, field2 = altFields

        # Base properties (all except alternates)
        baseProps = {k: v for k, v in properties.items() if k not in altFields}
        baseRequired = [f for f in required if f not in altFields]

        # Schema with field1 only
        schemaWithField1 = {
            "type": "object",
            "properties": {**baseProps, field1: properties[field1]},
            "additionalProperties": openApiSchemaDict.get(
                "additionalProperties", False
            ),
        }
        if field1 in required or baseRequired:
            schemaWithField1["required"] = baseRequired + (
                [field1] if field1 in required else []
            )
        oneOfSchemas.append(schemaWithField1)

        # Schema with field2 only
        schemaWithField2 = {
            "type": "object",
            "properties": {**baseProps, field2: properties[field2]},
            "additionalProperties": openApiSchemaDict.get(
                "additionalProperties", False
            ),
        }
        if field2 in required or baseRequired:
            schemaWithField2["required"] = baseRequired + (
                [field2] if field2 in required else []
            )
        oneOfSchemas.append(schemaWithField2)

    # Replace the schema with oneOf constraint
    if oneOfSchemas:
        openApiSchemaDict.clear()
        openApiSchemaDict["oneOf"] = oneOfSchemas


def enumSchemaFromNamedtuple(nt_instance, description=None):
    """
    Generate an OpenAPI schema dict for a namedtuple instance.
    """
    return {
        "type": "string",
        "enum": [v for v in nt_instance._asdict().values()],
        "description": description or nt_instance.__class__.__name__,
    }


def namedtupleToEnum(nt_instance, enum_name="AutoEnum"):
    if not isinstance(nt_instance, tuple) or not hasattr(nt_instance, "_fields"):
        raise TypeError("Must pass a namedtuple instance")

    # Build dict of {field_name: value}
    members = {field: getattr(nt_instance, field) for field in nt_instance._fields}

    # Dynamically create a subclass of str & Enum
    return Enum(enum_name, members, type=str)



# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.exchanging module

"""

import json
from typing import Union
from dataclasses import dataclass

import falcon
from keri import core
from keri.core import coring, eventing, serdering
from keri.peer import exchanging
from keri.help import ogler
from keria.core import httping
from keria.utils.openapi import dataclassFromFielddom

logger = ogler.getLogger()


def loadEnds(app):
    exnColEnd = ExchangeCollectionEnd()
    app.add_route("/identifiers/{name}/exchanges", exnColEnd)

    exnColEnd = ExchangeQueryCollectionEnd()
    app.add_route("/exchanges/query", exnColEnd)

    exnResEnd = ExchangeResourceEnd()
    app.add_route("/exchanges/{said}", exnResEnd)


exnFieldDomV1 = serdering.SerderKERI.Fields[serdering.Protocols.keri][
    serdering.Vrsn_1_0
][serdering.Ilks.exn]
EXN_V_1, EXNSchema_V_1 = dataclassFromFielddom("EXN_V_1", exnFieldDomV1)

exnFieldDomV2 = serdering.SerderKERI.Fields[serdering.Protocols.keri][
    serdering.Vrsn_2_0
][serdering.Ilks.exn]
EXN_V_2, EXNSchema_V_2 = dataclassFromFielddom("EXN_V_2", exnFieldDomV2)


class ExchangeCollectionEnd:
    @staticmethod
    def on_post(req, rep, name):
        """POST endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            name (str): human readable name or prefix for AID context
        ---
        summary: Post an exchange message for an identifier.
        description: This endpoint posts an exchange message to a specific named identifier.
        tags:
        - Exchange Message
        parameters:
        - in: path
          name: name
          schema:
            type: string
          required: true
          description: The human-readable alias for the AID context.
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    exn:
                      type: object
                      description: The exchange message event.
                    sigs:
                      type: array
                      items:
                        type: string
                      description: The signatures for the exn message.
                    atc:
                      type: object
                      description: The additional attachments for the exn message.
                    rec:
                      type: array
                      items:
                        type: string
                      description: The recipients of the exn message.
                    tpc:
                      type: string
                      description: The topic of the exn message.
        responses:
            202:
                description: Successfully posted the exchange message.
                content:
                    application/json:
                        schema:
                            type: object
                            $ref: '#/components/schemas/Exn'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: The requested identifier was not found.
        """
        agent = req.context.agent

        body = req.get_media()

        # Get the hab
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )
        if hab is None:
            raise falcon.HTTPNotFound(
                description=f"{name} is not a valid reference to an identifier"
            )

        # Get the exn, sigs, additional attachments and recipients  from the request
        ked = httping.getRequiredParam(body, "exn")
        sigs = httping.getRequiredParam(body, "sigs")
        atc = httping.getRequiredParam(body, "atc")
        rec = httping.getRequiredParam(body, "rec")
        topic = httping.getRequiredParam(body, "tpc")

        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(
                    description=f"attempt to send to unknown AID={recp}"
                )

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [core.Siger(qb64=sig) for sig in sigs]

        logger.info(
            "[%s | %s]: route %s received exn %s being sent to [%s]",
            name,
            hab.pre,
            serder.ked["r"],
            serder.said,
            ", ".join(rec),
        )

        # Now create the stream to send, need the signer seal
        kever = hab.kever
        seal = eventing.SealEvent(
            i=hab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d
        )

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)

        # Have to add the atc to the end... this will be Pathed signatures for embeds
        ims.extend(atc.encode("utf-8"))  # add the pathed attachments

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims), exc=agent.exc)

        msg = dict(said=serder.said, pre=hab.pre, rec=rec, topic=topic)

        logger.info("[%s | %s]: appending %s", name, hab.pre, json.dumps(msg))
        agent.exchanges.append(msg)

        rep.status = falcon.HTTP_202
        rep.data = json.dumps(serder.ked).encode("utf-8")


@dataclass
class ExchangeResource:
    """Data class for exchange message resource"""

    exn: Union["EXN_V_1", "EXN_V_2"]  # type: ignore
    pathed: dict

    def to_dict(self):
        return {
            "exn": self.exn,
            "pathed": {k: v.decode("utf-8") for k, v in self.pathed.items()},
        }


class ExchangeQueryCollectionEnd:
    @staticmethod
    def on_post(req, rep):
        """POST endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object

        ---
        summary: Query exchange message collection.
        description: This endpoint retrieves the exchange messages based on the provided query parameters.
        tags:
        - Exchange Message
        requestBody:
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    filter:
                      type: object
                      description: The filter criteria to apply on the exchange messages.
                    sort:
                      type: object
                      description: The sorting criteria to apply on the exchange messages.
                    skip:
                      type: integer
                      description: The number of exchange messages to skip. (default=0)
                    limit:
                      type: integer
                      description: The maximum number of exchange messages to return. (default=25)
        responses:
            200:
              description: Successfully retrieved the exchange messages.
              content:
                application/json:
                  schema:
                    type: array
                    items:
                      $ref: '#/components/schemas/ExchangeResource'
            400:
              description: Bad request. This could be due to missing or invalid parameters.
        """
        agent = req.context.agent

        try:
            body = req.get_media()
            if "filter" in body:
                filtr = body["filter"]
            else:
                filtr = {}

            if "sort" in body:
                sort = body["sort"]
            else:
                sort = None

            if "skip" in body:
                skip = body["skip"]
            else:
                skip = 0

            if "limit" in body:
                limit = body["limit"]
            else:
                limit = 25
        except falcon.HTTPError:
            filtr = {}
            sort = {}
            skip = 0
            limit = 25

        cur = agent.exnseeker.find(filtr=filtr, sort=sort, skip=skip, limit=limit)
        saids = [coring.Saider(qb64=said) for said in cur]

        exns = []
        for said in saids:
            serder, pathed = exchanging.cloneMessage(agent.hby, said.qb64)
            exns.append(
                dict(
                    exn=serder.ked,
                    pathed={k: v.decode("utf-8") for k, v in pathed.items()},
                )
            )

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(exns).encode("utf-8")


class ExchangeResourceEnd:
    """Exchange message resource endpoint class"""

    @staticmethod
    def on_get(req, rep, said):
        """GET endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            said (str): qb64 SAID of exchange message to retrieve

        ---
        summary: Retrieve a specific exchange message.
        description: This endpoint retrieves a specific exchange message based on the provided SAID.
        tags:
        - Exchange Message
        parameters:
        - in: path
          name: said
          schema:
            type: string
          required: true
          description: The qb64 SAID of the exchange message to retrieve.
        responses:
            200:
              description: Successfully retrieved the exchange message.
              content:
                application/json:
                  schema:
                    type: object
                    $ref: '#/components/schemas/ExchangeResource'
            404:
              description: The requested exchange message was not found.
        """
        agent = req.context.agent
        serder, pathed = exchanging.cloneMessage(agent.hby, said)

        if serder is None:
            raise falcon.HTTPNotFound(
                description=f"SAID {said} does not match a verified EXN message"
            )

        exn = dict(
            exn=serder.ked, pathed={k: v.decode("utf-8") for k, v in pathed.items()}
        )
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(exn).encode("utf-8")



# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.authing module

"""

from urllib.parse import quote, unquote
import falcon
from hio.help import Hict
from keri import kering
from keri.end import ending
from keri.help import helping


class Authenticater:
    DefaultFields = ["Signify-Resource", "@method", "@path", "Signify-Timestamp"]

    def __init__(self, agency):
        """Create Agent Authenticator for verifying requests and signing responses

        Parameters:
            agency(Agency): habitat of Agent for signing responses

        Returns:
              Authenicator:  the configured habery

        """
        self.agency = agency

    @staticmethod
    def resource(request):
        headers = request.headers
        if "SIGNIFY-RESOURCE" not in headers:
            raise ValueError("Missing signify resource header")

        return headers["SIGNIFY-RESOURCE"]

    def verify(self, request):
        headers = request.headers
        if "SIGNATURE-INPUT" not in headers or "SIGNATURE" not in headers:
            return False

        siginput = headers["SIGNATURE-INPUT"]
        if not siginput:
            return False
        signature = headers["SIGNATURE"]
        if not signature:
            return False

        inputs = ending.desiginput(siginput.encode("utf-8"))
        inputs = [i for i in inputs if i.name == "signify"]

        if not inputs:
            return False

        for inputage in inputs:
            items = []
            for field in inputage.fields:
                if field.startswith("@"):
                    if field == "@method":
                        items.append(f'"{field}": {request.method}')
                    elif field == "@path":
                        items.append(f'"{field}": {request.path}')

                else:
                    key = field.upper()
                    field = field.lower()
                    if key not in headers:
                        continue

                    value = ending.normalize(headers[key])
                    items.append(f'"{field}": {value}')

            values = [f"({' '.join(inputage.fields)})", f"created={inputage.created}"]
            if inputage.expires is not None:
                values.append(f"expires={inputage.expires}")
            if inputage.nonce is not None:
                values.append(f"nonce={inputage.nonce}")
            if inputage.keyid is not None:
                values.append(f"keyid={inputage.keyid}")
            if inputage.context is not None:
                values.append(f"context={inputage.context}")
            if inputage.alg is not None:
                values.append(f"alg={inputage.alg}")

            params = ";".join(values)

            items.append(f'"@signature-params: {params}"')
            ser = "\n".join(items).encode("utf-8")

            resource = self.resource(request)
            agent = self.agency.get(resource)

            if agent is None:
                raise kering.AuthNError("unknown or invalid controller")

            if resource not in agent.agentHab.kevers:
                raise kering.AuthNError("unknown or invalid controller")

            ckever = agent.agentHab.kevers[resource]
            signages = ending.designature(signature)
            cig = signages[0].markers[inputage.name]
            if not ckever.verfers[0].verify(sig=cig.raw, ser=ser):
                raise kering.AuthNError(f"Signature for {inputage} invalid")

        return True

    def sign(self, agent, headers, method, path, fields=None):
        """Generate and add Signature Input and Signature fields to headers

        Parameters:
            agent (Agent): The agent that is replying to the request
            headers (Hict): HTTP header to sign
            method (str): HTTP method name of request/response
            path (str): HTTP Query path of request/response
            fields (Optional[list]): Optional list of Signature Input fields to sign.

        Returns:
            headers (Hict): Modified headers with new Signature and Signature Input fields

        """

        if fields is None:
            fields = self.DefaultFields

        header, qsig = ending.siginput(
            "signify",
            method,
            path,
            headers,
            fields=fields,
            hab=agent.agentHab,
            alg="ed25519",
            keyid=agent.agentHab.pre,
        )
        headers.extend(header)
        signage = ending.Signage(
            markers=dict(signify=qsig),
            indexed=False,
            signer=None,
            ordinal=None,
            digest=None,
            kind=None,
        )
        headers.extend(ending.signature([signage]))

        return headers


class SignatureValidationComponent(object):
    """Validate Signature and Signature-Input header signatures"""

    def __init__(self, agency, authn: Authenticater, allowed=None):
        """

        Parameters:
            authn (Authenticater): Authenticator to validate signature headers on request
            allowed (list[str]): Paths that are not protected.
        """
        if allowed is None:
            allowed = []
        self.agency = agency
        self.authn = authn
        self.allowed = allowed

    def process_request(self, req, resp):
        """Process request to ensure has a valid signature from caid

        Parameters:
            req: Http request object
            resp: Http response object


        """

        for path in self.allowed:
            if req.path.startswith(path):
                return

        req.path = quote(req.path)

        try:
            # Use Authenticater to verify the signature on the request
            if self.authn.verify(req):
                req.path = unquote(req.path)
                resource = self.authn.resource(req)
                agent = self.agency.get(caid=resource)

                req.context.agent = agent
                return

        except kering.AuthNError:
            pass
        except ValueError:
            pass

        resp.complete = (
            True  # This short-circuits Falcon, skipping all further processing
        )
        resp.status = falcon.HTTP_401
        return

    def process_response(self, req, rep, resource, req_succeeded):
        """Process every falcon response by adding signature headers signed by the Agent AID.

        Parameters:
            req (Request): Falcon request object
            rep (Response): Falcon response object
            resource (End): endpoint object the request was routed to
            req_succeeded (boot): True means the request was successfully handled


        """

        if hasattr(req.context, "agent"):
            req.path = quote(req.path)
            agent = req.context.agent
            rep.set_header("Signify-Resource", agent.agentHab.pre)
            rep.set_header("Signify-Timestamp", helping.nowIso8601())
            headers = self.authn.sign(agent, Hict(rep.headers), req.method, req.path)
            for key, val in headers.items():
                rep.set_header(key, val)



import falcon

from hio.base import doing
from keri import kering, help
from keri.app import forwarding, agenting, habbing
from keri.core import coring, serdering
from keri.db import dbing
from keria.core import httping, longrunning

DELEGATION_ROUTE = "/identifiers/{name}/delegation"

logger = help.ogler.getLogger()


def loadEnds(app, identifierResource):
    gatorEnd = DelegatorEnd(identifierResource)
    app.add_route(DELEGATION_ROUTE, gatorEnd)


class Anchorer(doing.DoDoer):
    """
    Sends messages to Delegator of an identifier and wait for the anchoring event to
    be processed to ensure the inception or rotation event has been approved by the delegator.

    Removes all Doers and exits as Done once the event has been anchored.

    """

    def __init__(self, hby, proxy=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hby (Habery): Habery of the agent to populate witnesses
            proxy (Hab): Agent Hab to use as a messaging proxy to send messages to the delegator on behalf of SignifyHab instances as SignifyHabs cannot sign messages in a KERIA Agent.
        """
        self.hby = hby
        self.postman = forwarding.Poster(hby=hby)
        self.witq = agenting.WitnessInquisitor(hby=hby)
        self.witDoer = agenting.Receiptor(hby=self.hby)
        self.proxy = proxy

        super(Anchorer, self).__init__(
            doers=[self.witq, self.witDoer, self.postman, doing.doify(self.escrowDo)],
            **kwa,
        )

    def delegation(self, pre, sn=None, proxy=None):
        if pre not in self.hby.habs:
            raise kering.ValidationError(
                f"{pre} is not a valid local AID for delegation"
            )

        if proxy is not None:
            self.proxy = proxy

        # load the hab of the delegated identifier to anchor
        hab = self.hby.habs[pre]
        delpre = hab.kever.delpre  # get the delegator identifier
        if delpre not in hab.kevers:
            raise kering.ValidationError(
                f"delegator {delpre} not found, unable to process delegation"
            )

        sn = sn if sn is not None else hab.kever.sner.num

        # load the event and signatures
        evt = hab.makeOwnEvent(sn=sn)

        srdr = serdering.SerderKERI(raw=evt)
        self.witDoer.msgs.append(dict(pre=pre, sn=srdr.sn))
        self.hby.db.dpwe.pin(keys=(srdr.pre, srdr.said), val=srdr)

    def complete(self, prefixer, seqner, saider=None):
        """Check for completed delegation protocol for the specific event

        Parameters:
            prefixer (Prefixer): qb64 identifier prefix of event to check
            seqner (Seqner): sequence number of event to check
            saider (Saider): optional digest of event to verify

        Returns:

        """
        csaider = self.hby.db.cdel.get(keys=(prefixer.qb64, seqner.qb64))
        if not csaider:
            return False
        else:
            if saider and (csaider.qb64 != saider.qb64):
                raise kering.ValidationError(
                    f"invalid delegation protocol escrowed event {csaider.qb64}-{saider.qb64}"
                )

        return True

    def escrowDo(self, tymth, tock=1.0, temp=False, **opts):
        """Process escrows of group multisig identifiers waiting to be compeleted.

        Steps involve:
           1. Sending local event with sig to other participants
           2. Waiting for signature threshold to be met.
           3. If elected and delegated identifier, send complete event to delegator
           4. If delegated, wait for delegator's anchored seal
           5. If elected, send event to witnesses and collect receipts.
           6. Otherwise, wait for fully receipted event

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value.  Default to 1.0 to slow down processing

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = yield self.tock

        while True:
            self.processEscrows()
            yield 0.5

    def processEscrows(self):
        self.processPartialWitnessEscrow()
        self.processUnanchoredEscrow()

    def processUnanchoredEscrow(self):
        """
        Process escrow of unacnchored events that have been delegated and are waiting for delegator anchor/approval.
        """
        for (
            pre,
            said,
        ), serder in self.hby.db.dune.getItemIter():  # group partial witness escrow
            kever = self.hby.kevers[pre]
            dkever = self.hby.kevers[kever.delpre]

            seal = dict(i=serder.pre, s=serder.snh, d=serder.said)
            if dserder := self.hby.db.findAnchoringSealEvent(
                dkever.prefixer.qb64, seal=seal
            ):
                seqner = coring.Seqner(sn=dserder.sn)
                couple = seqner.qb64b + dserder.saidb
                dgkey = dbing.dgKey(kever.prefixer.qb64b, kever.serder.saidb)
                self.hby.db.setAes(
                    dgkey, couple
                )  # authorizer event seal (delegator/issuer)

                # Move to escrow waiting for witness receipts
                logger.info(
                    "[%s]: Delegation approval received, %s confirmed", pre, serder.pre
                )
                self.hby.db.cdel.put(
                    keys=(pre, coring.Seqner(sn=serder.sn).qb64),
                    val=coring.Saider(qb64=serder.said),
                )
                self.hby.db.dune.rem(keys=(pre, said))

    def processPartialWitnessEscrow(self):
        """
        Process escrow of delegated events that do not have a full compliment of receipts
        from witnesses yet.  When receipting is complete, remove from escrow and cue up a message
        that the event is complete.

        """
        for (
            pre,
            said,
        ), serder in self.hby.db.dpwe.getItemIter():  # group partial witness escrow
            hab = self.hby.habs[pre]
            kever = self.hby.kevers[pre]
            delpre = hab.kever.delpre  # get the delegator identifier
            dgkey = dbing.dgKey(pre, serder.said)
            seqner = coring.Seqner(sn=serder.sn)

            # Load all the witness receipts we have so far
            wigs = self.hby.db.getWigs(dgkey)
            if len(wigs) == len(
                kever.wits
            ):  # We have all of them, this event is finished
                if len(kever.wits) > 0:
                    witnessed = False
                    for cue in self.witDoer.cues:
                        if cue["pre"] == serder.pre and cue["sn"] == seqner.sn:
                            witnessed = True
                    if not witnessed:
                        continue
                logger.info(
                    "[%s]: Witness receipts complete, waiting for delegation approval for %s",
                    pre,
                    serder.pre,
                )
                if isinstance(hab, habbing.GroupHab):
                    phab = hab.mhab
                elif self.proxy is not None:
                    phab = self.proxy
                elif hab.kever.sn > 0:
                    phab = hab
                else:
                    raise kering.ValidationError(
                        "No proxy to send messages for delegation for AID %s]", pre
                    )

                evt = hab.db.cloneEvtMsg(pre=serder.pre, fn=0, dig=serder.said)

                srdr = serdering.SerderKERI(raw=evt)
                del evt[: srdr.size]
                self.postman.send(
                    hab=phab, dest=delpre, topic="delegate", serder=srdr, attachment=evt
                )

                self.hby.db.dpwe.rem(keys=(pre, said))
                self.hby.db.dune.pin(keys=(srdr.pre, srdr.said), val=srdr)


class DelegatorEnd:
    """Resource class for handling delegator events"""

    def __init__(self, identifierResource) -> None:
        """

        Parameters:
            identifierResource (IdentifierResourceEnd): endpoint class for creating rotation and interaction events

        """
        self.identifierResource = identifierResource

    def on_post(self, req, rep, name):
        """Identifier delegator endpoint POST to create the ixn anchor and approve the delegation

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object
            name (str): human readable name or prefix for Hab to rename

        responses:
            200:
                description: Delegation approved and interaction event created.
                content:
                  application/json:
                    schema:
                      - $ref: '#/components/schemas/DelegatorOperation'
            400:
                description: Bad request. This could be due to missing or invalid parameters.
            404:
                description: Not found. The specified identifier does not exist.

        """
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")
        agent = req.context.agent
        hab = (
            agent.hby.habs[name]
            if name in agent.hby.habs
            else agent.hby.habByName(name)
        )

        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID with name or prefix {name} found")

        body = req.get_media()
        anc = httping.getRequiredParam(body, "ixn")

        if not agent.hby.db.findAnchoringSealEvent(hab.pre, seal=anc):
            op = self.identifierResource.interact(agent, name, body)

        # successful approval returns the delegatee prefix
        teepre = approveDelegation(hab, anc)
        adop = agent.monitor.submit(
            anc["d"],
            longrunning.OpTypes.delegation,
            metadata=dict(
                pre=hab.kever.prefixer.qb64, teepre=teepre, anchor=anc, depends=op
            ),
        )

        try:
            rep.status = falcon.HTTP_200
            rep.content_type = "application/json"
            rep.data = adop.to_json().encode("utf-8")
            return rep
        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])


def approveDelegation(hab, anc) -> str:
    serder = serdering.SerderKERI(sad=anc)

    teepre = anc["a"][0]["i"]
    teesaid = anc["a"][0]["d"]

    for (pre, sn), dig in hab.db.delegables.getItemIter():
        if pre == teepre:
            seqner = coring.Seqner(sn=serder.sn)
            couple = seqner.qb64b + serder.saidb
            dgkey = dbing.dgKey(
                coring.Saider(qb64=teepre).qb64b, coring.Saider(qb64=teesaid).qb64b
            )
            # the dip event should have been received from the delegatee via a postman call
            # and will be sitting in the delegator escrows (hence the hab.db.delegables above)
            # adding the authorize event seal will allow the dip to be processed
            # and added to the delegator kever
            hab.db.setAes(dgkey, couple)  # authorizer event seal (delegator/issuer)

    return teepre
    # raise falcon.HTTPBadRequest(title=f"No delegables found for delegator {hab.pre} to approve delegatee {teepre}")



import signal

from hio.base import doing
from keri import help

logger = help.ogler.getLogger()


class GracefulShutdownDoer(doing.Doer):
    """
    Shuts all Agency agents down before throwing a KeyboardInterrup to cause the Doist loop to exit,
    performing a graceful shutdown.
    Sets up signal handler in the Doer.enter lifecycle method.
    Checks for the shutdown flag being set in the Doer.recur lifecycle method.
    """

    def __init__(self, agency, **kwa):
        """
        Sets up the GracefulShutdownDoer with the Agency. Eventually the Agency should manage itself
        with its own exit function that shuts down all agents gracefully.

        Parameters:
            agency (Agency): The Agency containing Agent instances to be gracefully shut down
            kwa (dict): Additional keyword arguments for Doer initialization
        """
        self.agency = agency
        self.shutdown_received = False

        super().__init__(**kwa)

    def handle_sigterm(self, signum, frame):
        """Handler function for SIGTERM"""
        logger.info("Received SIGTERM, initiating graceful shutdown.")
        self.shutdown_received = True

    def handle_sigint(self, signum, frame):
        """Handler function for SIGINT"""
        logger.info("Received SIGINT, initiating graceful shutdown.")
        self.shutdown_received = True

    def shutdownAgency(self):
        """Helper function to shut down the agents."""
        logger.info("Stopping Agency")
        self.agency.shouldShutdown = True  # Set the shutdown flag in the Agency

    def enter(self, temp=False):
        """
        Sets up signal handlers.
        Lifecycle method called once when the Doist running this Doer enters the context for this Doer.
        """
        # Register signal handler
        signal.signal(signal.SIGTERM, self.handle_sigterm)
        signal.signal(signal.SIGINT, self.handle_sigint)
        logger.info("Registered signal handlers for SIGTERM")

    def recur(self, tyme=None, tock=0.0, **opts):
        """Generator coroutine checking once per tock for shutdown flag"""
        # Checks once per tock if the shutdown flag has been set and if so initiates the shutdown process
        while not self.shutdown_received:
            yield tock  # will iterate forever in here until shutdown flag set
        logger.info("Shutdown flag received, initiating graceful shutdown of agents")
        self.shutdownAgency()
        # Once shutdown_received is set, trigger agency shutdown which will eventually shut down
        # the Doist loop by throwing a KeyboardInterrupt
        return True  # Returns a "done" status



import copy
import falcon
import marshmallow_dataclass
from apispec import yaml_utils
from apispec.core import VALID_METHODS, APISpec
from apispec.ext.marshmallow import MarshmallowPlugin

from keria.app import aiding, agenting, grouping, notifying
from keria.peer import exchanging
from ..core import optypes
from ..utils.openapi import applyAltConstraintsToOpenApiSchema
from . import credentialing
from keri.core import coring
from ..utils.openapi import enumSchemaFromNamedtuple

"""
KERIA
keria.app.specing module

OpenAPI Description Resource for the KERI and ACDC ReST interface
"""


class AgentSpecResource:
    """
    OpenAPI Description Resource for the KERI and ACDC ReST interface

    Contains all the endpoint descriptions for the KERI admin interface including:
    1. Creating and managing autonomic identifiers (AIDs) including multi-signature groups.
    2. Creating and managing authentic chained data containers (ACDCs)
    """

    def __init__(self, app, title, version="1.0.1", openapi_version="3.1.0"):
        self.spec = APISpec(
            title=title,
            version=version,
            openapi_version=openapi_version,
            plugins=[MarshmallowPlugin()],
        )

        # Register marshmallow schemas (pass class)
        self.spec.components.schema("ACDC_V_1", schema=credentialing.ACDCSchema_V_1)
        self.spec.components.schema("ACDC_V_2", schema=credentialing.ACDCSchema_V_2)
        self.spec.components.schema(
            "IssEvent",
            schema=marshmallow_dataclass.class_schema(credentialing.IssEvent)(),
        )
        self.spec.components.schema(
            "Schema", schema=marshmallow_dataclass.class_schema(credentialing.Schema)()
        )
        self.spec.components.schema(
            "Anchor", schema=marshmallow_dataclass.class_schema(credentialing.Anchor)()
        )
        self.spec.components.schema(
            "Seal", schema=marshmallow_dataclass.class_schema(aiding.Seal)()
        )
        self.spec.components.schema(
            "IXN_V_1",
            schema=marshmallow_dataclass.class_schema(aiding.IXN_V_1)(),
        )
        self.spec.components.schema(
            "IXN_V_2",
            schema=marshmallow_dataclass.class_schema(aiding.IXN_V_2)(),
        )
        self.spec.components.schema(
            "ICP_V_1",
            schema=marshmallow_dataclass.class_schema(aiding.ICP_V_1)(),
        )
        self.spec.components.schema(
            "ICP_V_2",
            schema=marshmallow_dataclass.class_schema(aiding.ICP_V_2)(),
        )
        self.spec.components.schema(
            "ROT_V_1",
            schema=marshmallow_dataclass.class_schema(aiding.ROT_V_1)(),
        )
        self.spec.components.schema(
            "ROT_V_2",
            schema=marshmallow_dataclass.class_schema(aiding.ROT_V_2)(),
        )
        self.spec.components.schema(
            "DIP_V_1",
            schema=marshmallow_dataclass.class_schema(aiding.DIP_V_1)(),
        )
        self.spec.components.schema(
            "DIP_V_2",
            schema=marshmallow_dataclass.class_schema(aiding.DIP_V_2)(),
        )
        self.spec.components.schema(
            "DRT_V_1",
            schema=marshmallow_dataclass.class_schema(aiding.DRT_V_1)(),
        )
        self.spec.components.schema(
            "DRT_V_2",
            schema=marshmallow_dataclass.class_schema(aiding.DRT_V_2)(),
        )
        self.spec.components.schema(
            "RPY_V_1", schema=marshmallow_dataclass.class_schema(aiding.RPY_V_1)()
        )
        self.spec.components.schema(
            "RPY_V_2", schema=marshmallow_dataclass.class_schema(aiding.RPY_V_2)()
        )
        self.spec.components.schema(
            "VCP_V_1",
            schema=marshmallow_dataclass.class_schema(agenting.VCP_V_1)(),
        )
        self.spec.components.schema(
            "ISS_V_1",
            schema=marshmallow_dataclass.class_schema(agenting.ISS_V_1)(),
        )
        self.spec.components.schema(
            "REV_V_1",
            schema=marshmallow_dataclass.class_schema(agenting.REV_V_1)(),
        )
        self.spec.components.schema(
            "EXN_V_1",
            schema=marshmallow_dataclass.class_schema(exchanging.EXN_V_1)(),
        )
        self.spec.components.schema(
            "EXN_V_2",
            schema=marshmallow_dataclass.class_schema(exchanging.EXN_V_2)(),
        )
        self.spec.components.schema(
            "Credential",
            schema=marshmallow_dataclass.class_schema(credentialing.ClonedCredential)(),
        )
        self.spec.components.schema(
            "CredentialStateIssOrRev",
            schema=marshmallow_dataclass.class_schema(
                credentialing.CredentialStateIssOrRev
            ),
        )
        self.spec.components.schema(
            "CredentialStateBisOrBrv",
            schema=marshmallow_dataclass.class_schema(
                credentialing.CredentialStateBisOrBrv
            ),
        )

        # Patch the schema to force additionalProperties=True
        acdc_attributes_schema = self.spec.components.schemas["ACDCAttributes"]
        acdc_attributes_schema["additionalProperties"] = True

        # The ACDC class has alts constraints like {'a': 'A', 'A': 'a'}
        acdc_schema_v1 = self.spec.components.schemas["ACDC_V_1"]
        if hasattr(credentialing.ACDCSchema_V_1, "_alt_constraints"):
            applyAltConstraintsToOpenApiSchema(
                acdc_schema_v1, credentialing.ACDCSchema_V_1._alt_constraints
            )

        acdc_schema_v2 = self.spec.components.schemas["ACDC_V_2"]
        if hasattr(credentialing.ACDCSchema_V_2, "_alt_constraints"):
            applyAltConstraintsToOpenApiSchema(
                acdc_schema_v2, credentialing.ACDCSchema_V_2._alt_constraints
            )

        credentialSchema = self.spec.components.schemas["Credential"]
        credentialSchema["properties"]["sad"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ACDC_V_1"},
                {"$ref": "#/components/schemas/ACDC_V_2"},
            ]
        }
        ancEvent = {
            "oneOf": [
                {"$ref": "#/components/schemas/IXN_V_1"},
                {"$ref": "#/components/schemas/IXN_V_2"},
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/DIP_V_1"},
                {"$ref": "#/components/schemas/DIP_V_2"},
                {"$ref": "#/components/schemas/DRT_V_1"},
                {"$ref": "#/components/schemas/DRT_V_2"},
            ]
        }
        credentialSchema["properties"]["anc"] = ancEvent

        # CredentialState
        self.spec.components.schemas["CredentialState"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/CredentialStateIssOrRev"},
                {"$ref": "#/components/schemas/CredentialStateBisOrBrv"},
            ]
        }

        credentialSchema["properties"]["status"] = {
            "$ref": "#/components/schemas/CredentialState"
        }

        # Registries
        self.spec.components.schema(
            "Registry",
            schema=marshmallow_dataclass.class_schema(credentialing.Registry)(),
        )
        registrySchema = self.spec.components.schemas["Registry"]
        registrySchema["properties"]["state"] = {
            "$ref": "#/components/schemas/CredentialState"
        }

        self.spec.components.schema(
            "AgentResourceResult",
            schema=marshmallow_dataclass.class_schema(aiding.AgentResourceResult)(),
        )

        agentControllerSchema = self.spec.components.schemas["Controller"]
        agentControllerSchema["properties"]["ee"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/DIP_V_1"},
                {"$ref": "#/components/schemas/DIP_V_2"},
                {"$ref": "#/components/schemas/DRT_V_1"},
                {"$ref": "#/components/schemas/DRT_V_2"},
            ]
        }

        # Register HabState
        self.spec.components.schema(
            "HabState", schema=marshmallow_dataclass.class_schema(aiding.HabState)()
        )
        self.spec.components.schema(
            "SaltyState", schema=marshmallow_dataclass.class_schema(aiding.SaltyState)()
        )
        self.spec.components.schema(
            "RandyKeyState",
            schema=marshmallow_dataclass.class_schema(aiding.RandyKeyState)(),
        )
        self.spec.components.schema(
            "GroupKeyState",
            schema=marshmallow_dataclass.class_schema(aiding.GroupKeyState)(),
        )
        self.spec.components.schema(
            "ExternState",
            schema=marshmallow_dataclass.class_schema(aiding.ExternState)(),
        )

        statesList = [
            {
                "required": ["salty"],
                "properties": {"salty": {"$ref": "#/components/schemas/SaltyState"}},
            },
            {
                "required": ["randy"],
                "properties": {"randy": {"$ref": "#/components/schemas/RandyKeyState"}},
            },
            {
                "required": ["group"],
                "properties": {"group": {"$ref": "#/components/schemas/GroupKeyState"}},
            },
            {
                "required": ["extern"],
                "properties": {"extern": {"$ref": "#/components/schemas/ExternState"}},
            },
        ]

        self.spec.components.schema(
            "HabStateBase",
            schema=marshmallow_dataclass.class_schema(aiding.HabStateBase)(),
        )
        habStateSchemaBase = self.spec.components.schemas["HabStateBase"]
        habStateSchemaBase["oneOf"] = statesList
        habStateSchema = self.spec.components.schemas["HabState"]
        habStateSchema["oneOf"] = statesList

        self.spec.components.schemas["GroupKeyState"]["properties"]["mhab"] = {
            "$ref": "#/components/schemas/HabState"
        }
        self.spec.components.schemas["Tier"] = enumSchemaFromNamedtuple(
            coring.Tiers, description="Tier of key material"
        )
        saltyStateSchema = self.spec.components.schemas["SaltyState"]
        saltyStateSchema["properties"]["tier"] = {"$ref": "#/components/schemas/Tier"}

        # Patch the schema to force additionalProperties=True
        externStateSchema = self.spec.components.schemas["ExternState"]
        externStateSchema["additionalProperties"] = True

        # OOBIS
        self.spec.components.schema(
            "OOBI", schema=marshmallow_dataclass.class_schema(aiding.OOBI)()
        )

        # End Roles
        self.spec.components.schema(
            "EndRole", schema=marshmallow_dataclass.class_schema(aiding.EndRole)()
        )

        # Register the Challenge schema
        self.spec.components.schema(
            "Challenge", schema=marshmallow_dataclass.class_schema(aiding.Challenge)()
        )

        # Register the Contact schema
        self.spec.components.schema(
            "Contact", schema=marshmallow_dataclass.class_schema(aiding.Contact)()
        )
        # Patch the schema to force additionalProperties=True
        contactSchema = self.spec.components.schemas["Contact"]
        contactSchema["additionalProperties"] = True

        # Register the GroupMember schema
        self.spec.components.schema(
            "GroupMember",
            schema=marshmallow_dataclass.class_schema(aiding.GroupMember)(),
        )

        # Register the KeyEventRecord schema
        self.spec.components.schema(
            "KeyEventRecord",
            schema=marshmallow_dataclass.class_schema(agenting.KeyEventRecord)(),
        )
        keyEventRecordSchema = self.spec.components.schemas["KeyEventRecord"]
        keyEventRecordSchema["properties"]["ked"] = ancEvent

        # Register the AgentConfig schema
        self.spec.components.schema(
            "AgentConfig",
            schema=marshmallow_dataclass.class_schema(agenting.AgentConfig)(),
        )

        # Register mulisig exn schemas
        self.spec.components.schemas["Exn"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/EXN_V_1"},
                {"$ref": "#/components/schemas/EXN_V_2"},
            ]
        }
        self.spec.components.schemas["Icp"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
            ]
        }
        self.spec.components.schemas["Rot"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
            ]
        }
        self.spec.components.schemas["Vcp"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/VCP_V_1"},
            ]
        }
        self.spec.components.schemas["Iss"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ISS_V_1"},
            ]
        }
        self.spec.components.schemas["Ixn"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/IXN_V_1"},
                {"$ref": "#/components/schemas/IXN_V_2"},
            ]
        }
        self.spec.components.schemas["Rpy"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/RPY_V_1"},
                {"$ref": "#/components/schemas/RPY_V_2"},
            ]
        }

        # Register Notification schema
        self.spec.components.schema(
            "Notification",
            schema=marshmallow_dataclass.class_schema(notifying.Notification)(),
        )

        # Patch the schema to force additionalProperties=True
        notificationDataSchema = self.spec.components.schemas["NotificationData"]
        notificationDataSchema["additionalProperties"] = True

        # Register the ExchangeResource schema
        self.spec.components.schema(
            "ExchangeResource",
            schema=marshmallow_dataclass.class_schema(exchanging.ExchangeResource)(),
        )
        exchangeResourceSchema = self.spec.components.schemas["ExchangeResource"]
        exchangeResourceSchema["properties"]["exn"] = {
            "$ref": "#/components/schemas/Exn"
        }

        # Register Grouping schemas
        self.spec.components.schema(
            "MultisigInceptEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigInceptEmbeds),
        )
        self.spec.components.schemas["MultisigInceptEmbeds"]["properties"]["icp"] = {
            "$ref": "#/components/schemas/Icp"
        }

        self.spec.components.schema(
            "MultisigRotateEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigRotateEmbeds),
        )
        self.spec.components.schemas["MultisigRotateEmbeds"]["properties"]["rot"] = {
            "$ref": "#/components/schemas/Rot"
        }

        self.spec.components.schema(
            "MultisigInteractEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigInteractEmbeds),
        )
        self.spec.components.schemas["MultisigInteractEmbeds"]["properties"]["ixn"] = {
            "$ref": "#/components/schemas/Ixn"
        }

        self.spec.components.schema(
            "MultisigRegistryInceptEmbeds",
            schema=marshmallow_dataclass.class_schema(
                grouping.MultisigRegistryInceptEmbeds
            ),
        )
        self.spec.components.schemas["MultisigRegistryInceptEmbeds"]["properties"][
            "vcp"
        ] = {"$ref": "#/components/schemas/Vcp"}
        self.spec.components.schemas["MultisigRegistryInceptEmbeds"]["properties"][
            "anc"
        ] = ancEvent

        self.spec.components.schema(
            "MultisigIssueEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigIssueEmbeds),
        )
        self.spec.components.schemas["MultisigIssueEmbeds"]["properties"]["acdc"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ACDC_V_1"},
                {"$ref": "#/components/schemas/ACDC_V_2"},
            ]
        }
        self.spec.components.schemas["MultisigIssueEmbeds"]["properties"]["iss"] = {
            "$ref": "#/components/schemas/Iss"
        }
        self.spec.components.schemas["MultisigIssueEmbeds"]["properties"]["anc"] = (
            ancEvent
        )

        self.spec.components.schema(
            "MultisigRevokeEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigRevokeEmbeds),
        )
        self.spec.components.schemas["MultisigRevokeEmbeds"]["properties"]["rev"] = {
            "$ref": "#/components/schemas/REV_V_1"
        }
        self.spec.components.schemas["MultisigRevokeEmbeds"]["properties"]["anc"] = (
            ancEvent
        )

        self.spec.components.schema(
            "MultisigRpyEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigRpyEmbeds),
        )
        self.spec.components.schemas["MultisigRpyEmbeds"]["properties"]["rpy"] = {
            "$ref": "#/components/schemas/Rpy"
        }

        self.spec.components.schema(
            "MultisigExnEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.MultisigExnEmbeds),
        )
        self.spec.components.schemas["MultisigExnEmbeds"]["properties"]["exn"] = {
            "$ref": "#/components/schemas/Exn"
        }

        self.spec.components.schema(
            "ExnEmbeds",
            schema=marshmallow_dataclass.class_schema(grouping.ExnEmbedsBase)(),
        )
        exnEmbedsSchema = self.spec.components.schemas["ExnEmbeds"]
        exnEmbedsSchema["oneOf"] = [
            {"$ref": "#/components/schemas/MultisigInceptEmbeds"},
            {"$ref": "#/components/schemas/MultisigRotateEmbeds"},
            {"$ref": "#/components/schemas/MultisigInteractEmbeds"},
            {"$ref": "#/components/schemas/MultisigRegistryInceptEmbeds"},
            {"$ref": "#/components/schemas/MultisigIssueEmbeds"},
            {"$ref": "#/components/schemas/MultisigRevokeEmbeds"},
            {"$ref": "#/components/schemas/MultisigRpyEmbeds"},
            {"$ref": "#/components/schemas/MultisigExnEmbeds"},
        ]
        self.spec.components.schema(
            "ExnMultisig",
            schema=marshmallow_dataclass.class_schema(grouping.ExnMultisig)(),
        )
        exnMSchema = self.spec.components.schemas["ExnMultisig"]
        exnMSchema["properties"]["exn"] = {"$ref": "#/components/schemas/Exn"}

        # Patch KeyStateRecord
        keyStateRecordSchema = self.spec.components.schemas["KeyStateRecord"]
        keyStateRecordSchema["properties"]["kt"] = {
            "oneOf": [
                {"type": "string"},
                {"type": "array", "items": {"type": "string"}},
            ]
        }
        keyStateRecordSchema["properties"]["nt"] = {
            "oneOf": [
                {"type": "string"},
                {"type": "array", "items": {"type": "string"}},
            ]
        }
        if "kt" not in keyStateRecordSchema["required"]:
            keyStateRecordSchema["required"].append("kt")
        if "nt" not in keyStateRecordSchema["required"]:
            keyStateRecordSchema["required"].append("nt")

        # Register the OOBI operation schemas
        self.spec.components.schema(
            "PendingOOBIOperation",
            schema=marshmallow_dataclass.class_schema(optypes.PendingOOBIOperation)(),
        )
        self.spec.components.schema(
            "CompletedOOBIOperation",
            schema=marshmallow_dataclass.class_schema(optypes.CompletedOOBIOperation)(),
        )
        self.spec.components.schema(
            "FailedOOBIOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedOOBIOperation)(),
        )
        self.spec.components.schemas["OOBIOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingOOBIOperation"},
                {"$ref": "#/components/schemas/CompletedOOBIOperation"},
                {"$ref": "#/components/schemas/FailedOOBIOperation"},
            ]
        }

        # Register the Query operation schemas
        self.spec.components.schema(
            "PendingQueryOperation",
            schema=marshmallow_dataclass.class_schema(optypes.PendingQueryOperation)(),
        )
        self.spec.components.schema(
            "CompletedQueryOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedQueryOperation
            )(),
        )
        self.spec.components.schema(
            "FailedQueryOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedQueryOperation)(),
        )
        self.spec.components.schemas["QueryOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingQueryOperation"},
                {"$ref": "#/components/schemas/CompletedQueryOperation"},
                {"$ref": "#/components/schemas/FailedQueryOperation"},
            ]
        }

        # Register the EndRole operation schemas
        self.spec.components.schema(
            "PendingEndRoleOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingEndRoleOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedEndRoleOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedEndRoleOperation
            )(),
        )
        self.spec.components.schema(
            "FailedEndRoleOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedEndRoleOperation)(),
        )
        self.spec.components.schemas["CompletedEndRoleOperation"]["properties"][
            "response"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/RPY_V_1"},
                {"$ref": "#/components/schemas/RPY_V_2"},
            ]
        }
        self.spec.components.schemas["EndRoleOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingEndRoleOperation"},
                {"$ref": "#/components/schemas/CompletedEndRoleOperation"},
                {"$ref": "#/components/schemas/FailedEndRoleOperation"},
            ]
        }

        # Register the Witness operation schemas
        self.spec.components.schema(
            "PendingWitnessOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingWitnessOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedWitnessOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedWitnessOperation
            )(),
        )
        self.spec.components.schema(
            "FailedWitnessOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedWitnessOperation)(),
        )
        self.spec.components.schemas["CompletedWitnessOperation"]["properties"][
            "response"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/IXN_V_1"},
                {"$ref": "#/components/schemas/IXN_V_2"},
            ]
        }
        self.spec.components.schemas["WitnessOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingWitnessOperation"},
                {"$ref": "#/components/schemas/CompletedWitnessOperation"},
                {"$ref": "#/components/schemas/FailedWitnessOperation"},
            ]
        }

        # Register the Delegation operation schemas
        self.spec.components.schema(
            "PendingDelegationOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingDelegationOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedDelegationOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedDelegationOperation
            )(),
        )
        self.spec.components.schema(
            "FailedDelegationOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedDelegationOperation
            )(),
        )
        self.spec.components.schemas["CompletedDelegationOperation"]["properties"][
            "response"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/DIP_V_1"},
                {"$ref": "#/components/schemas/DIP_V_2"},
                {"$ref": "#/components/schemas/DRT_V_1"},
                {"$ref": "#/components/schemas/DRT_V_2"},
            ]
        }
        self.spec.components.schemas["DelegationOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingDelegationOperation"},
                {"$ref": "#/components/schemas/CompletedDelegationOperation"},
                {"$ref": "#/components/schemas/FailedDelegationOperation"},
            ]
        }

        # Registry operation schemas
        self.spec.components.schema(
            "PendingRegistryOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingRegistryOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedRegistryOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedRegistryOperation
            )(),
        )
        self.spec.components.schema(
            "FailedRegistryOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedRegistryOperation
            )(),
        )
        self.spec.components.schemas["RegistryOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingRegistryOperation"},
                {"$ref": "#/components/schemas/CompletedRegistryOperation"},
                {"$ref": "#/components/schemas/FailedRegistryOperation"},
            ]
        }

        # LocScheme operation schemas
        self.spec.components.schema(
            "PendingLocSchemeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingLocSchemeOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedLocSchemeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedLocSchemeOperation
            )(),
        )
        self.spec.components.schema(
            "FailedLocSchemeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedLocSchemeOperation
            )(),
        )
        self.spec.components.schemas["LocSchemeOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingLocSchemeOperation"},
                {"$ref": "#/components/schemas/CompletedLocSchemeOperation"},
                {"$ref": "#/components/schemas/FailedLocSchemeOperation"},
            ]
        }

        # Challenge operation schemas
        self.spec.components.schema(
            "PendingChallengeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingChallengeOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedChallengeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedChallengeOperation
            )(),
        )
        self.spec.components.schema(
            "FailedChallengeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedChallengeOperation
            )(),
        )
        self.spec.components.schemas["ChallengeOperationResponse"]["properties"][
            "exn"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/EXN_V_1"},
                {"$ref": "#/components/schemas/EXN_V_2"},
            ]
        }
        self.spec.components.schemas["ChallengeOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingChallengeOperation"},
                {"$ref": "#/components/schemas/CompletedChallengeOperation"},
                {"$ref": "#/components/schemas/FailedChallengeOperation"},
            ]
        }

        # Exchange operation schemas
        self.spec.components.schema(
            "PendingExchangeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingExchangeOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedExchangeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedExchangeOperation
            )(),
        )
        self.spec.components.schema(
            "FailedExchangeOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedExchangeOperation
            )(),
        )
        self.spec.components.schemas["ExchangeOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingExchangeOperation"},
                {"$ref": "#/components/schemas/CompletedExchangeOperation"},
                {"$ref": "#/components/schemas/FailedExchangeOperation"},
            ]
        }

        # Submit operation schemas
        self.spec.components.schema(
            "PendingSubmitOperation",
            schema=marshmallow_dataclass.class_schema(optypes.PendingSubmitOperation)(),
        )
        self.spec.components.schema(
            "CompletedSubmitOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedSubmitOperation
            )(),
        )
        self.spec.components.schema(
            "FailedSubmitOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedSubmitOperation)(),
        )
        self.spec.components.schemas["SubmitOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingSubmitOperation"},
                {"$ref": "#/components/schemas/CompletedSubmitOperation"},
                {"$ref": "#/components/schemas/FailedSubmitOperation"},
            ]
        }

        # Done operation schemas
        self.spec.components.schema(
            "DoneOperationMetadata",
            schema=marshmallow_dataclass.class_schema(optypes.DoneOperationMetadata)(),
        )
        self.spec.components.schemas["DoneOperationMetadata"]["properties"][
            "response"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/EXN_V_1"},
                {"$ref": "#/components/schemas/EXN_V_2"},
            ]
        }
        self.spec.components.schema(
            "PendingDoneOperation",
            schema=marshmallow_dataclass.class_schema(optypes.PendingDoneOperation)(),
        )
        self.spec.components.schema(
            "CompletedDoneOperation",
            schema=marshmallow_dataclass.class_schema(optypes.CompletedDoneOperation)(),
        )
        self.spec.components.schema(
            "FailedDoneOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedDoneOperation)(),
        )
        self.spec.components.schemas["CompletedDoneOperation"]["properties"][
            "response"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ICP_V_1"},
                {"$ref": "#/components/schemas/ICP_V_2"},
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/EXN_V_1"},
                {"$ref": "#/components/schemas/EXN_V_2"},
            ]
        }
        self.spec.components.schemas["DoneOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingDoneOperation"},
                {"$ref": "#/components/schemas/CompletedDoneOperation"},
                {"$ref": "#/components/schemas/FailedDoneOperation"},
            ]
        }

        # Credential operation schemas
        self.spec.components.schema(
            "CredentialOperationMetadata",
            schema=marshmallow_dataclass.class_schema(
                optypes.CredentialOperationMetadata
            )(),
        )
        self.spec.components.schemas["CredentialOperationMetadata"]["properties"][
            "ced"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ACDC_V_1"},
                {"$ref": "#/components/schemas/ACDC_V_2"},
            ]
        }
        self.spec.components.schemas["CredentialOperationMetadata"]["properties"][
            "depends"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ROT_V_1"},
                {"$ref": "#/components/schemas/ROT_V_2"},
                {"$ref": "#/components/schemas/DRT_V_1"},
                {"$ref": "#/components/schemas/DRT_V_2"},
                {"$ref": "#/components/schemas/IXN_V_1"},
                {"$ref": "#/components/schemas/IXN_V_2"},
            ]
        }
        self.spec.components.schema(
            "CredentialOperationResponse",
            schema=marshmallow_dataclass.class_schema(
                optypes.CredentialOperationResponse
            )(),
        )
        self.spec.components.schemas["CredentialOperationResponse"]["properties"][
            "ced"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/ACDC_V_1"},
                {"$ref": "#/components/schemas/ACDC_V_2"},
            ]
        }
        self.spec.components.schema(
            "PendingCredentialOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingCredentialOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedCredentialOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedCredentialOperation
            )(),
        )
        self.spec.components.schema(
            "FailedCredentialOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedCredentialOperation
            )(),
        )
        self.spec.components.schemas["CredentialOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingCredentialOperation"},
                {"$ref": "#/components/schemas/CompletedCredentialOperation"},
                {"$ref": "#/components/schemas/FailedCredentialOperation"},
            ]
        }

        # Group operation schemas
        self.spec.components.schema(
            "PendingGroupOperation",
            schema=marshmallow_dataclass.class_schema(optypes.PendingGroupOperation)(),
        )
        self.spec.components.schema(
            "CompletedGroupOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedGroupOperation
            )(),
        )
        self.spec.components.schema(
            "FailedGroupOperation",
            schema=marshmallow_dataclass.class_schema(optypes.FailedGroupOperation)(),
        )
        self.spec.components.schemas["CompletedGroupOperation"]["properties"][
            "response"
        ] = ancEvent
        self.spec.components.schemas["GroupOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingGroupOperation"},
                {"$ref": "#/components/schemas/CompletedGroupOperation"},
                {"$ref": "#/components/schemas/FailedGroupOperation"},
            ]
        }

        # Delegator operation schemas
        self.spec.components.schema(
            "DelegatorOperationMetadata",
            schema=marshmallow_dataclass.class_schema(
                optypes.DelegatorOperationMetadata
            )(),
        )
        self.spec.components.schemas["DelegatorOperationMetadata"]["properties"][
            "depends"
        ] = {
            "oneOf": [
                {"$ref": "#/components/schemas/GroupOperation"},
                {"$ref": "#/components/schemas/WitnessOperation"},
                {"$ref": "#/components/schemas/DoneOperation"},
            ]
        }

        self.spec.components.schema(
            "PendingDelegatorOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.PendingDelegatorOperation
            )(),
        )
        self.spec.components.schema(
            "CompletedDelegatorOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.CompletedDelegatorOperation
            )(),
        )
        self.spec.components.schema(
            "FailedDelegatorOperation",
            schema=marshmallow_dataclass.class_schema(
                optypes.FailedDelegatorOperation
            )(),
        )
        self.spec.components.schemas["DelegatorOperation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/PendingDelegatorOperation"},
                {"$ref": "#/components/schemas/CompletedDelegatorOperation"},
                {"$ref": "#/components/schemas/FailedDelegatorOperation"},
            ]
        }

        self.spec.components.schemas["Operation"] = {
            "oneOf": [
                {"$ref": "#/components/schemas/OOBIOperation"},
                {"$ref": "#/components/schemas/QueryOperation"},
                {"$ref": "#/components/schemas/EndRoleOperation"},
                {"$ref": "#/components/schemas/WitnessOperation"},
                {"$ref": "#/components/schemas/DelegationOperation"},
                {"$ref": "#/components/schemas/RegistryOperation"},
                {"$ref": "#/components/schemas/LocSchemeOperation"},
                {"$ref": "#/components/schemas/ChallengeOperation"},
                {"$ref": "#/components/schemas/ExchangeOperation"},
                {"$ref": "#/components/schemas/SubmitOperation"},
                {"$ref": "#/components/schemas/DoneOperation"},
                {"$ref": "#/components/schemas/CredentialOperation"},
                {"$ref": "#/components/schemas/GroupOperation"},
                {"$ref": "#/components/schemas/DelegatorOperation"},
            ]
        }

        self.addRoutes(app)

    def addRoutes(self, app):
        valid_methods = self._get_valid_methods(self.spec)
        routes_to_check = copy.copy(app._router._roots)

        for route in routes_to_check:
            if route.resource is not None:
                operations = dict()
                operations.update(
                    yaml_utils.load_operations_from_docstring(route.resource.__doc__)
                    or {}
                )

                if route.method_map:
                    for method_name, method_handler in route.method_map.items():
                        if method_handler.__module__ == "falcon.responders":
                            continue
                        if method_name.lower() not in valid_methods:
                            continue
                        docstring_yaml = yaml_utils.load_yaml_from_docstring(
                            method_handler.__doc__
                        )
                        operations[method_name.lower()] = docstring_yaml or dict()

                self.spec.path(path=route.uri_template, operations=operations)
            routes_to_check.extend(route.children)

    def _get_valid_methods(self, spec):
        return set(VALID_METHODS[spec.openapi_version.major])

    def on_get(self, _, rep):
        """
        GET endpoint for OpenAPI 3.1.0 spec

        Args:
            _: falcon.Request HTTP request
            rep: falcon.Response HTTP response


        """
        rep.status = falcon.HTTP_200
        rep.content_type = "application/yaml"
        rep.data = self._get_spec_yaml()

    def _get_spec_yaml(self):
        return self.spec.to_yaml().encode("utf-8")



# -*- encoding: utf-8 -*-
"""
KERIA
keria.end.ending module

ReST API endpoints

"""

import falcon
from keri import kering
from keri.end import ending


def loadEnds(app, agency, default=None):
    end = OOBIEnd(agency=agency, default=default)
    app.add_route("/oobi", end)
    app.add_route("/oobi/{aid}", end)
    app.add_route("/oobi/{aid}/{role}", end)
    app.add_route("/oobi/{aid}/{role}/{eid}", end)


class OOBIEnd:
    """REST API for OOBI endpoints

    Attributes:
        .hby (Habery): database access

    """

    def __init__(self, agency, default=None):
        """End point for responding to OOBIs

        Parameters:
            default (str) qb64 AID of the 'self' of the node for

        """
        self.agency = agency
        self.default = default

    def on_get(self, _, rep, aid=None, role=None, eid=None):
        """GET endoint for OOBI resource

        Parameters:
            _: Falcon request object
            rep: Falcon response object
            aid: qb64 identifier prefix of OOBI
            role: requested role for OOBI rpy message
            eid: qb64 identifier prefix of participant in role
        ---
        summary: Retrieve OOBI resource.
        description: This endpoint retrieves the OOBI resource based on the provided aid, role, and eid.
        tags:
        - OOBI Resource
        parameters:
        - in: path
          name: aid
          schema:
            type: string
          required: false
          description: The qb64 identifier prefix of OOBI.
        - in: path
          name: role
          schema:
            type: string
          required: false
          description: The requested role for OOBI rpy message.
        - in: path
          name: eid
          schema:
            type: string
          required: false
          description: The qb64 identifier prefix of participant in role.
        responses:
            200:
                description: Successfully retrieved the OOBI resource.
            400:
                description: Bad request. This could be due to invalid or missing parameters.
            404:
                description: The requested OOBI resource was not found.
        """
        if not aid:
            if self.default is None:
                raise falcon.HTTPNotFound(description="no blind oobi for this node")

            aid = self.default

        agent = self.agency.lookup(pre=aid)
        if agent is None:
            raise falcon.HTTPNotFound(description="AID not found for this OOBI")

        kever = agent.hby.kevers[aid]
        if not agent.hby.db.fullyWitnessed(kever.serder):
            raise falcon.HTTPNotFound(description=f"{aid} not available")

        if kever.prefixer.qb64 in agent.hby.prefixes:  # One of our identifiers
            hab = agent.hby.habs[kever.prefixer.qb64]
        else:  # Not allowed to respond
            raise falcon.HTTPNotAcceptable(
                description=f"{aid} is not a local identifier"
            )

        eids = []
        if eid:
            eids.append(eid)

        msgs = hab.replyToOobi(aid=aid, role=role, eids=eids)
        if not msgs and role is None:
            msgs = hab.replyToOobi(aid=aid, role=kering.Roles.witness, eids=eids)
            msgs.extend(hab.replay(aid))

        if msgs:
            rep.status = falcon.HTTP_200  # This is the default status
            rep.set_header(ending.OOBI_AID_HEADER, aid)
            rep.content_type = "application/json+cesr"
            rep.data = bytes(msgs)

        else:
            rep.status = falcon.HTTP_NOT_FOUND



# -*- encoding: utf-8 -*-
"""
KERIA
keria.db.basing module

"""

from dataclasses import dataclass
from ordered_set import OrderedSet as oset

from keri.core import coring
from keri.db import dbing, subing, koming

SCALAR_TYPES = ("string", "number")

ISSUER_FIELD = coring.Pather(path=["i"])
ISSUEE_FIELD = coring.Pather(path=["a", "i"])
SCHEMA_FIELD = coring.Pather(path=["s"])
REGISTRY_FIELD = coring.Pather(path=["ri"])


@dataclass
class IndexRecord:
    """Registry Key keyed by Registry name"""

    subkey: str
    paths: list


class AgencyBaser(dbing.LMDBer):
    """
    Agency database for tracking Agent tenants and their managed identifiers in this KERIA instance.

    """

    TailDirPath = "keri/adb"
    AltTailDirPath = ".keri/adb"
    TempPrefix = "keri_adb_"
    MaxNamedDBs = 10

    def __init__(self, headDirPath=None, perm=None, reopen=False, **kwa):
        """
        Setup named sub databases.

        Inherited Parameters:
            name is str directory path name differentiator for main database
                When system employs more than one keri database, name allows
                differentiating each instance by name
                default name='main'
            temp is boolean, assign to .temp
                True then open in temporary directory, clear on close
                Otherwise then open persistent directory, do not clear on close
                default temp=False
            headDirPath is optional str head directory pathname for main database
                If not provided use default .HeadDirpath
                default headDirPath=None so uses self.HeadDirPath
            perm is numeric optional os dir permissions mode
                default perm=None so do not set mode
            reopen is boolean, IF True then database will be reopened by this init
                default reopen=True

        Attributes:
            .agnt values are Prefixer of Agent AID
                keyed by controller AID (caid).
                Maps Controller AIDs to their Signify Agent AID (agent.pre).
            .ctrl values are Prefixer of Controller AID (caid)
                keyed by Agent AID.
                Maps Agent AIDs to their Signify Controller AID (caid).
            .aids values are Prefixer of Controller AID (caid)
                keyed by managed AID.
                Maps managed AIDs to their Signify Controller AID (caid).

        Notes:
            dupsort=True for sub DB means allow unique (key,pair) duplicates at a key.
            Duplicate means that is more than one value at a key but not a redundant
            copies a (key,value) pair per key. In other words the pair (key,value)
            must be unique both key and value in combination.
            Attempting to put the same (key,value) pair a second time does
            not add another copy.

            Duplicates are inserted in lexicographic order by value, insertion order.
        """
        if perm is None:
            perm = self.Perm  # defaults to restricted permissions for non temp

        self.agnt = None
        self.ctrl = None
        self.aids = None

        super(AgencyBaser, self).__init__(
            headDirPath=headDirPath, perm=perm, reopen=reopen, **kwa
        )

    def reopen(self, **kwa):
        """
        Open sub databases
        """
        self.opened = super(AgencyBaser, self).reopen(**kwa)

        # Create by opening first time named sub DBs within main DB instance
        # Names end with "." as sub DB name must include a non Base64 character
        # to avoid namespace collisions with Base64 identifier prefixes.

        # Sub-database keyed by qb64 controller AID mapping to the Prefixer object of the AID of an agent
        self.agnt = subing.CesrSuber(db=self, subkey="agnt.", klas=coring.Prefixer)

        # Sub-database keyed by qb64 agent AID mapping to the Prefixer object of the AID of an controller
        self.ctrl = subing.CesrSuber(db=self, subkey="ctrl.", klas=coring.Prefixer)

        # Sub-database keyed by qb64 AID mapping to the Prefixer object of the AID of its Agent
        self.aids = subing.CesrSuber(db=self, subkey="aids.", klas=coring.Prefixer)


class Seeker(dbing.LMDBer):
    """
    Seeker indexes all credentials in the KERIpy `saved` Creder database.

    Static indexes are created for issued by/schema and issued to/schema and dynamic indexes are
    created for all top level scalar valued fields in the credential payload ('a' field).  The Seeker
    uses the schema of each credential to determine the payload fields to index by.

    """

    TailDirPath = "keri/seekdb"
    AltTailDirPath = ".keri/seekdb"
    TempPrefix = "keri_seekdb_"
    MaxNamedDBs = 500

    def __init__(self, db, reger, headDirPath=None, perm=None, reopen=False, **kwa):
        """
        Setup named sub databases.

        Inherited Parameters:
            name is str directory path name differentiator for main database
                When system employs more than one keri database, name allows
                differentiating each instance by name
                default name='main'
            temp is boolean, assign to .temp
                True then open in temporary directory, clear on close
                Othewise then open persistent directory, do not clear on close
                default temp=False
            headDirPath is optional str head directory pathname for main database
                If not provided use default .HeadDirpath
                default headDirPath=None so uses self.HeadDirPath
            perm is numeric optional os dir permissions mode
                default perm=None so do not set mode
            reopen is boolean, IF True then database will be reopened by this init
                default reopen=True


        """
        self.db = db
        self.reger = reger
        self.indexes = dict()

        self.schIdx = None
        self.dynIdx = None

        super(Seeker, self).__init__(
            headDirPath=headDirPath, perm=perm, reopen=reopen, **kwa
        )

    def reopen(self, **kwa):
        super(Seeker, self).reopen(**kwa)

        # List of indexs for a given schema
        self.schIdx = subing.IoSetSuber(db=self, subkey="schIdx.")
        # List of dynamically created indexes to be recreated at load
        self.dynIdx = koming.Komer(
            db=self,
            subkey="dynIdx.",
            schema=IndexRecord,
        )

        for name, idx in self.dynIdx.getItemIter():
            key = ".".join(name)
            self.indexes[key] = subing.CesrDupSuber(
                db=self, subkey=idx.subkey, klas=coring.Saider
            )

        # Create persistent Indexes if they don't already exist
        self.createIndex(SCHEMA_FIELD.qb64)
        self.createIndex(REGISTRY_FIELD.qb64)

        # Index of credentials by issuer/issuee.
        for field in (ISSUER_FIELD, ISSUEE_FIELD):
            self.createIndex(field.qb64)
            subkey = f"{field.qb64}.{SCHEMA_FIELD.qb64}"
            self.createIndex(subkey)

    @property
    def table(self):
        return self.reger.saved

    def value(self, said):
        saider = self.reger.saved.get(keys=(said,))
        creder = self.reger.creds.get(keys=(saider.qb64,))
        return creder.sad

    def saidIter(self):
        return self.reger.saved.getItemIter()

    def createIndex(self, key):
        if self.dynIdx.get(keys=(key,)) is None:
            self.indexes[key] = subing.CesrDupSuber(
                db=self, subkey=key, klas=coring.Saider
            )
            self.dynIdx.pin(keys=(key,), val=IndexRecord(subkey=key, paths=[key]))

    def index(self, said):
        if (saider := self.reger.saved.get(keys=(said,))) is None:
            raise ValueError(f"{said} is not a verified credential")

        creder = self.reger.creds.get(keys=(saider.qb64,))
        saider = coring.Saider(qb64b=creder.saidb)

        # Load schema index and if not indexed in schIdx, index it.
        schemaSaid = creder.schema
        if not (indexes := self.schIdx.get(keys=(schemaSaid,))):
            indexes = self.generateIndexes(schemaSaid)

        for index in indexes:
            db = self.indexes[index]
            idx = self.dynIdx.get(keys=(index,))
            values = []
            for path in idx.paths:
                pather = coring.Pather(qb64=path)
                values.append(pather.resolve(creder.sad))

            value = "".join(values)
            db.add(keys=(value,), val=saider)

    def unindex(self, said):
        if (saider := self.reger.saved.get(keys=(said,))) is None:
            raise ValueError(f"{said} is not a verified credential")

        creder = self.reger.creds.get(keys=(saider.qb64,))
        indexes = self.schIdx.get(keys=(creder.schema,))
        if not indexes:
            raise ValueError(f"No known indexes for schema {creder.schema}")

        for index in indexes:
            db = self.indexes[index]
            idx = self.dynIdx.get(keys=(index,))
            values = []
            for path in idx.paths:
                pather = coring.Pather(qb64=path)
                values.append(pather.resolve(creder.sad))

            value = "".join(values)
            db.rem(keys=(value,), val=saider)

    def generateIndexes(self, said):
        """Parse schema of said, create schIdx entry keyed to said of schema and the subkey indexes in
        self.indexes

        """
        if (schemer := self.db.schema.get(keys=(said,))) is None:
            raise ValueError(f"{said} is not a valid schema SAID to index")

        properties = schemer.sed["properties"]
        payload = properties["a"]
        if "oneOf" in payload:
            oo = payload["oneOf"]
            for p in oo:
                if p["type"] == "object":
                    payload = p
                    break

        if not isinstance(payload, dict):
            raise ValueError(
                f"schema with SAID={said} is not value, no payload of values"
            )

        properties = payload["properties"]

        attestation = "i" not in properties

        # Assign single field Schema and ISSUER index and ISSUER/SCHEMA:
        self.schIdx.add(keys=(said,), val=SCHEMA_FIELD.qb64b)
        self.schIdx.add(keys=(said,), val=ISSUER_FIELD.qb64b)
        self.schIdx.add(keys=(said,), val=REGISTRY_FIELD.qb64b)
        subkey = f"{ISSUER_FIELD.qb64}.{SCHEMA_FIELD.qb64}"
        self.schIdx.add(keys=(said,), val=subkey.encode("UTF-8"))

        # Assign single field ISSUEE index and ISSUEE/SCHEMA if needed
        if not attestation:
            self.schIdx.add(keys=(said,), val=ISSUEE_FIELD.qb64b)
            subkey = f"{ISSUEE_FIELD.qb64}.{SCHEMA_FIELD.qb64}"
            self.schIdx.add(keys=(said,), val=subkey.encode("UTF-8"))

        for p, val in properties.items():
            if val["type"] not in SCALAR_TYPES:
                continue

            pather = coring.Pather(path=["a", p])
            if pather.qb64 not in self.indexes:
                self.indexes[pather.qb64] = subing.CesrDupSuber(
                    db=self, subkey=pather.qb64, klas=coring.Saider
                )
                idx = IndexRecord(subkey=pather.qb64, paths=[pather.qb64])
                self.dynIdx.pin(keys=(pather.qb64,), val=idx)
            self.schIdx.add(keys=(said,), val=pather.qb64b)

            subkey = f"{SCHEMA_FIELD.qb64}.{pather.qb64}"
            if subkey not in self.indexes:
                self.indexes[subkey] = subing.CesrDupSuber(
                    db=self, subkey=subkey, klas=coring.Saider
                )
                idx = IndexRecord(subkey=subkey, paths=[SCHEMA_FIELD.qb64, pather.qb64])
                self.dynIdx.pin(keys=(subkey,), val=idx)
            self.schIdx.add(keys=(said,), val=subkey)

            for field in (ISSUER_FIELD, ISSUEE_FIELD):
                subkey = f"{field.qb64}.{pather.qb64}"
                if subkey not in self.indexes:
                    self.indexes[subkey] = subing.CesrDupSuber(
                        db=self, subkey=subkey, klas=coring.Saider
                    )
                    idx = IndexRecord(subkey=subkey, paths=[field.qb64, pather.qb64])
                    self.dynIdx.pin(keys=(subkey,), val=idx)

                self.schIdx.add(keys=(said,), val=subkey)

                subkey = f"{field.qb64}.{SCHEMA_FIELD.qb64}.{pather.qb64}"
                if subkey not in self.indexes:
                    self.indexes[subkey] = subing.CesrDupSuber(
                        db=self, subkey=subkey, klas=coring.Saider
                    )
                    idx = IndexRecord(
                        subkey=subkey,
                        paths=[field.qb64, SCHEMA_FIELD.qb64, pather.qb64],
                    )
                    self.dynIdx.pin(keys=(subkey,), val=idx)

                self.schIdx.add(keys=(said,), val=subkey)

        return [index for index in self.schIdx.get(keys=(said,))]

    def find(self, filtr, sort=None, skip=None, limit=None):
        return Cursor(seeker=self, filtr=filtr, sort=sort, skip=skip, limit=limit)


class ExnSeeker(dbing.LMDBer):
    """
    Seeker indexes all credentials in the KERIpy `saved` Creder database.

    Static indexes are created for issued by/schema and issued to/schema and dynamic indexes are
    created for all top level scalar valued fields in the credential payload ('a' field).  The Seeker
    uses the schema of each credential to determine the payload fields to index by.

    """

    TailDirPath = "keri/exndb"
    AltTailDirPath = ".keri/exndb"
    TempPrefix = "keri_exndb_"
    MaxNamedDBs = 36

    DATE_FIELD = coring.Pather(path=["dt"])
    SENDER_FIELD = coring.Pather(path=["i"])
    RECIPIENT_FIELD = coring.Pather(path=["a", "i"])
    ROUTE_FIELD = coring.Pather(path=["r"])

    # Special field for IPEX messages... consider moving to IpexSeeker if needed
    SCHEMA = coring.Pather(path=["e", "acdc", "s"])

    def __init__(self, db, headDirPath=None, perm=None, reopen=False, **kwa):
        """
        Setup named sub databases.

        Inherited Parameters:
            name is str directory path name differentiator for main database
                When system employs more than one keri database, name allows
                differentiating each instance by name
                default name='main'
            temp is boolean, assign to .temp
                True then open in temporary directory, clear on close
                Othewise then open persistent directory, do not clear on close
                default temp=False
            headDirPath is optional str head directory pathname for main database
                If not provided use default .HeadDirpath
                default headDirPath=None so uses self.HeadDirPath
            perm is numeric optional os dir permissions mode
                default perm=None so do not set mode
            reopen is boolean, IF True then database will be reopened by this init
                default reopen=True


        """
        self.db = db
        self.indexes = dict()

        super(ExnSeeker, self).__init__(
            headDirPath=headDirPath, perm=perm, reopen=reopen, **kwa
        )

    def reopen(self, **kwa):
        super(ExnSeeker, self).reopen(**kwa)

        # List of dynamically created indexes to be recreated at load
        # Create persistent Indexes if they don't already exist
        fields = (
            self.ROUTE_FIELD,
            self.SENDER_FIELD,
            self.RECIPIENT_FIELD,
            self.DATE_FIELD,
            self.SCHEMA,
        )
        # Index of credentials by issuer/issuee.
        for field in fields:
            self.createIndex(field.qb64)
            for subfield in fields:
                if field == subfield:
                    continue
                subkey = f"{field.qb64}.{subfield.qb64}"
                self.createIndex(subkey)

    @property
    def table(self):
        return self.db.exns

    def value(self, said):
        serder = self.db.exns.get(keys=(said,))
        return serder.ked

    def saidIter(self):
        for (said,), _ in self.db.exns.getItemIter():
            yield said

    def createIndex(self, key):
        self.indexes[key] = subing.CesrDupSuber(db=self, subkey=key, klas=coring.Saider)

    def index(self, said):
        if (serder := self.db.exns.get(keys=(said,))) is None:
            raise ValueError(f"{said} is not a valid exn")

        saider = coring.Saider(qb64b=serder.saidb)

        # Load schema index and if not indexed in schIdx, index it.
        for index, db in self.indexes.items():
            pathers = [coring.Pather(qb64=path) for path in index.split(".")]
            values = []
            for pather in pathers:
                try:
                    values.append(pather.resolve(serder.ked))
                except KeyError:
                    pass

            value = "".join(values)
            if not value:
                continue

            db.add(keys=(value,), val=saider)

    def find(self, filtr, sort=None, skip=None, limit=None):
        return Cursor(seeker=self, filtr=filtr, sort=sort, skip=skip, limit=limit)


class Cursor:
    def __init__(self, seeker, filtr=None, sort=None, skip=None, limit=None):
        self.filtr = filtr
        self.operators = operators(self.filtr)
        self.names = [op.name for op in self.operators]
        self.indexable = next(
            (False for op in self.operators if not isinstance(op, Eq)), True
        )
        self.values = [op.value for op in self.operators]

        self.seeker = seeker
        self._sort = sort
        self._skip = skip if skip is not None else 0
        self._limit = limit if limit is not None else 25

        self.cur = None
        self.saids = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.saids is None:
            self._query()

        if self.cur >= len(self.saids):
            raise StopIteration

        said = self.saids[self.cur]
        self.cur += 1
        return said

    def sort(self, sort):
        self._sort = sort
        return self

    def skip(self, skip):
        self._skip = skip
        return self

    def limit(self, limit):
        self._limit = limit
        return self

    def _query(self):
        self.cur = 0
        if len(self.filtr) == 0:
            self.saids = self.order(
                [said for (said,), _ in self.seeker.table.getItemIter()]
            )
        elif (saids := self.indexSearch()) is not None:
            self.saids = self.order(saids)
        elif (saids := self.indexScan()) is not None:
            self.saids = self.order(saids)
        else:
            saids = self.fullTableScan()
            self.saids = self.order(saids)

    def indexSearch(self):
        if len(self.operators) == 1 and self.operators[0].name in self.seeker.indexes:
            op = self.operators[0]
            idx = self.seeker.indexes[op.name]
            return op.index(idx)

        index = ".".join(self.names)
        if not (self.indexable and index in self.seeker.indexes):
            return None

        idx = self.seeker.indexes[index]
        val = "".join(self.values)
        return [val.qb64 for val in idx.getIter(keys=(val,))]

    def indexScan(self):
        use = []
        scan = []

        for idx, op in enumerate(self.operators):
            if op.name in self.seeker.indexes:
                use.append(op)
            else:
                scan.append(op)

        if len(use) == 0:
            return self.fullTableScan()

        idx = self.seeker.indexes[use[0].name]
        saids = oset(use[0].index(idx))
        if len(saids) == 0:
            return list()

        for i, op in enumerate(use[1:]):
            idx = self.seeker.indexes[op.name]
            nxt = oset(op.index(idx))
            saids &= nxt

        if len(scan) == 0:
            return list(saids)
        else:
            return self.tableScan(list(saids), scan)

    def fullTableScan(self):
        saids = [saider.qb64 for _, saider in self.seeker.saidIter()]
        return self.tableScan(saids, ops=self.operators)

    def tableScan(self, saids, ops):
        res = []
        for said in saids:
            val = self.seeker.value(said)
            for op in ops:
                if op(val):
                    res.append(said)

        return res

    def order(self, saids):
        if not self._sort:
            return self.slice(saids)

        if (res := self.indexOrder(saids)) is not None:
            return res
        else:
            return self.tableScanOrder(saids)

    def indexOrder(self, saids):
        index = ".".join([coring.Pather(bext=s).qb64 for s in self._sort])
        if index not in self.seeker.indexes:
            return None

        idx = self.seeker.indexes[index]
        ctx = idx.getItemIter()

        # Run off the values before start
        cur = 0
        while cur < self._skip:
            _, saider = next(ctx)
            if saider.qb64 in saids:
                cur += 1

        # Load values in order until we get enough (based on limit)
        res = []
        for _, saider in ctx:
            if saider.qb64 in saids:
                res.append(saider.qb64)

            if len(res) == self._limit:
                break

        return res

    def slice(self, saids):
        if self._skip >= len(saids):
            return []

        end = self._skip + self._limit
        return saids[self._skip : end]

    def tableScanOrder(self, saids):
        """Should we bother implementing table scan sort order

        We have single field indexes for all fields in credentials so this situation will
        only occur if multiple fields are selected for which we don't have a multi-column
        index.  In that case, perhaps we raise an exception.

        For now, we'll just slice the results to honor skip and limit.

        """
        return self.slice(saids)


def operators(filtr):
    """Executable operator factory method

    An factory for processing a filter dict and generating an array of
    executable operators to apply to a given credential search

    """
    # filtr = {"-a-i": {"$begins": "984"}}
    ops = []
    for f, v in filtr.items():
        if isinstance(v, dict):
            for op, val in v.items():
                match op:
                    case "$eq":
                        ops.append(Eq(field=f, value=val))
                    case "$begins":
                        ops.append(Begins(field=f, value=val))
            pass
        else:
            ops.append(Eq(field=f, value=v))

    return ops


class Eq:
    def __init__(self, field, value):
        self.field = field
        self.pather = coring.Pather(bext=self.field)
        self.value = value

    def __call__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError(
                f"invalid argument length={len(args)} for equals operator, must be 2"
            )

        val = self.pather.resolve(args[0])
        return val == self.value

    @property
    def name(self) -> str:
        return self.pather.qb64

    def index(self, idx):
        return [val.qb64 for val in idx.getIter(keys=(self.value,))]


class Begins:
    def __init__(self, field, value):
        self.field = field
        self.pather = coring.Pather(bext=self.field)

        if not isinstance(value, str):
            raise ValueError(f"invalid type={type(value)} for begins, must be `str`")
        self.value = value

    def __call__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError(
                f"invalid argument length={len(args)} for begins operator, must be 2"
            )

        val = self.pather.resolve(args[0])
        if not isinstance(val, str):
            raise ValueError(f"invalid type={type(args[0])} for begins, must be `str`")

        return val.startswith(self.value)

    def index(self, idx):
        return [val.qb64 for _, val in idx.getItemIter(keys=(self.value,))]

    @property
    def name(self) -> str:
        return self.pather.qb64
